"""Meta-algorithms to build online regression models."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

from skpro.regression.online._dont_refit import OnlineDontRefit
from skpro.regression.online._refit import OnlineRefit
from skpro.regression.online._refit_every import OnlineRefitEveryN

__all__ = ["OnlineDontRefit", "OnlineRefit", "OnlineRefitEveryN"]



# LEGACY MODULE - TODO: remove or refactor

import hashlib
import os
import urllib.request

import numpy as np
from sklearn.datasets import load_diabetes
from sklearn.model_selection import train_test_split


def retrieve_data(url):
    file_name = "/tmp/%s.dat" % hashlib.md5(str(url).encode("utf-8")).hexdigest()
    if not os.path.isfile(file_name):
        with urllib.request.urlopen(url) as response, open(file_name, "wb") as out_file:
            data = response.read()
            out_file.write(data)

    return file_name


def load_file(file, return_X_y=False):
    data = np.loadtxt(file)

    if return_X_y:
        return data[:, :-1], data[:, -1:].ravel()
    else:
        return data


def load_url(url, return_X_y=False):
    file = retrieve_data(url)
    return load_file(file, return_X_y)


class DataManager:
    """
    A helper to manage datasets more easily. Test/training split
    is carried out behind the scenes whenever new data is being
    assigned

    Parameters
    ----------
    X : np.array | string
        Features or 'boston', 'diabetes' to load sklearn datasets, url of file
    y : np.array
        Labels
    split: float, default=0.2
        Train/test split
    name: string, default=None
        Optional name to be used in the object representation
    random_state: int, default=None
        Optional random state to be used during split

    Attributes
    ----------
    X_train : np.array
        Training features
    X_test : np.array
        Training labels
    y_train : np.array
        Test features
    y_test : np.array
        Test labels
    """

    def __init__(self, X=None, y=None, split=0.2, name=None, random_state=None):
        if isinstance(X, str):
            # autoload sklearn datasets, urls and files
            name = X
            if name.lower() == "boston":
                X, y = load_diabetes(return_X_y=True)
            elif name.lower() == "diabetes":
                X, y = load_diabetes(return_X_y=True)
            elif name.startswith("http"):
                X, y = load_url(name, return_X_y=True)
            elif os.path.isfile(name):
                X, y = load_file(name, return_X_y=True)
            else:
                raise ValueError("'%s' is not a valid dataset" % X)

        self.random_state = random_state
        self.split = split
        self._X = None
        self._y = None
        self.X_train = None
        self.X_test = None
        self.y_train = None
        self.y_test = None
        self.X = X
        self.y = y
        self.name = name if isinstance(name, str) else "Unnamed"

    def data(self, copy=True):
        """Returns the data

        Parameters
        ----------
        copy: boolean, default=True
            If false, reference copy will be used

        Returns
        -------
        X, y
        """
        if not copy:
            return self.X, self.y

        try:
            X_ = np.copy(self._X)
            y_ = np.copy(self._y)
            return X_, y_
        except:
            return False

    def clone(self):
        """Clones the data manager

        Returns
        -------
        A copy of the data manager itself
        """
        X, y = self.data()
        return DataManager(X, y, self.split, name=self.name)

    def _split(self):
        if not isinstance(self.split, float):
            return

        if self._X is None or self._y is None:
            return

        if len(self._X) != len(self._y):
            return

        self.X_train, self.X_test, self.y_train, self.y_test = train_test_split(
            self._X, self._y, test_size=self.split, random_state=self.random_state
        )

    def shuffle(self, random_state=None):
        """Shuffles the data

        Parameters
        ----------
        random_state: int, default=None
            Optional random state

        Returns
        -------
        None
        """
        if random_state is not None:
            self.random_state = random_state

        data = list(zip(np.copy(self.X), np.copy(self.y)))
        np.random.shuffle(data)
        X, y = zip(*data)

        self.X, self.y = np.array(X), np.array(y)

    @property
    def X(self):
        return self._X

    @X.setter
    def X(self, value):
        self._X = value
        self._split()

    @property
    def y(self):
        return self._y

    @y.setter
    def y(self, value):
        self._y = value
        self._split()



"""Testing machine type checkers for scitypes."""

__author__ = ["fkiraly"]

import numpy as np
import pytest

from skpro.datatypes._check import (
    AMBIGUOUS_MTYPES,
    check_is_mtype,
    check_is_scitype,
    get_check_dict,
)
from skpro.datatypes._check import mtype as infer_mtype
from skpro.datatypes._check import scitype as infer_scitype
from skpro.datatypes._examples import get_examples
from skpro.datatypes._registry import SCITYPE_LIST, scitype_to_mtype
from skpro.tests.test_switch import run_test_module_changed

SCITYPES = SCITYPE_LIST

# scitypes where mtype inference is not unique
# alignment is excluded since mtypes can be ambiguous
#   (indices could be both loc or iloc when integers)
SCITYPES_AMBIGUOUS_MTYPE = []


def _generate_scitype_mtype_combinations():
    """Return scitype/mtype tuples for pytest_generate_tests.

    Fixtures parameterized
    ----------------------
    scitype : str - scitype of fixture
    mtype : str - mtype of fixture
    """
    # collect fixture tuples here

    sci_mtype_tuples = []

    for scitype in SCITYPES:
        mtypes = scitype_to_mtype(scitype)

        for mtype in mtypes:
            sci_mtype_tuples += [(scitype, mtype)]

    return sci_mtype_tuples


def _generate_scitype_mtype_fixtureindex_combinations():
    """Return fixture tuples for pytest_generate_tests.

    Fixtures parameterized
    ----------------------
    scitype : str - scitype of fixture
    mtype : str - mtype of fixture
    fixture_index : int - index of fixture tuple with that scitype and mtype
    """
    # collect fixture tuples here

    sci_mtype_tuples = _generate_scitype_mtype_combinations()

    sci_mtype_index_tuples = []

    for tuple_j in sci_mtype_tuples:
        scitype = tuple_j[0]
        mtype = tuple_j[1]
        n_fixtures = len(get_examples(mtype=mtype, as_scitype=scitype))

        for i in range(n_fixtures):
            sci_mtype_index_tuples += [(scitype, mtype, i)]

    return sci_mtype_index_tuples


def pytest_generate_tests(metafunc):
    """Test parameterization routine for pytest.

    Fixtures parameterized
    ----------------------
    scitype : str - scitype of fixture
    mtype : str - mtype of fixture
    fixture_index : int - index of fixture tuple with that scitype and mtype
    """
    # we assume all four arguments are present in the test below

    fixturenames = set(metafunc.fixturenames)

    if {"scitype", "mtype", "fixture_index"}.issubset(fixturenames):
        keys = _generate_scitype_mtype_fixtureindex_combinations()

        ids = []
        for tuple in keys:
            ids += [f"{tuple[0]}-{tuple[1]}-fixture:{tuple[2]}"]

        # parameterize test with from-mtpes
        metafunc.parametrize("scitype,mtype,fixture_index", keys, ids=ids)

    elif {"scitype", "mtype"}.issubset(fixturenames):
        keys = _generate_scitype_mtype_combinations()

        ids = []
        for tuple in keys:
            ids += [f"{tuple[0]}-{tuple[1]}"]

        # parameterize test with from-mtpes
        metafunc.parametrize("scitype,mtype", keys, ids=ids)


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes"),
    reason="Test only if skpro.datatypes has been changed",
)
def test_check_positive(scitype, mtype, fixture_index):
    """Tests that check_is_mtype correctly confirms the mtype of examples.

    Parameters
    ----------
    scitype : str - scitype of fixture
    mtype : str - mtype of fixture
    fixture_index : int - index of fixture tuple with that scitype and mtype

    Raises
    ------
    RuntimeError if scitype is not defined or has no mtypes or examples
    AssertionError if examples are not correctly identified
    error if check itself raises an error
    """
    # retrieve fixture for checking
    fixture = get_examples(mtype=mtype, as_scitype=scitype).get(fixture_index)

    # todo: possibly remove this once all checks are defined
    check_dict = get_check_dict()
    check_is_defined = (mtype, scitype) in check_dict.keys()

    # check fixtures that exist against checks that exist, when full metadata is queried
    if fixture is not None and check_is_defined:
        check_result = check_is_mtype(
            fixture, mtype, scitype, return_metadata=True, msg_return_dict="list"
        )
        if not check_result[0]:
            msg = (
                f"check_is_mtype returns False on scitype {scitype}, mtype {mtype} "
                f"fixture {fixture_index}, message: "
            )
            msg = msg + check_result[1]
        assert check_result[0], msg

    # check fixtures that exist against checks that exist, when no metadata is queried
    if fixture is not None and check_is_defined:
        check_result = check_is_mtype(
            fixture, mtype, scitype, return_metadata=[], msg_return_dict="list"
        )
        if not check_result[0]:
            msg = (
                f"check_is_mtype returns False on scitype {scitype}, mtype {mtype} "
                f"fixture {fixture_index}, message: "
            )
            msg = msg + check_result[1]
        assert check_result[0], msg


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes"),
    reason="Test only if skpro.datatypes has been changed",
)
def test_check_positive_check_scitype(scitype, mtype, fixture_index):
    """Tests that check_is_scitype correctly confirms the scitype of examples.

    Parameters
    ----------
    scitype : str - scitype of fixture
    mtype : str - mtype of fixture
    fixture_index : int - index of fixture tuple with that scitype and mtype

    Raises
    ------
    RuntimeError if scitype is not defined or has no mtypes or examples
    AssertionError if examples are not correctly identified
    error if check itself raises an error
    """
    if mtype in AMBIGUOUS_MTYPES:
        return None

    # retrieve fixture for checking
    fixture = get_examples(mtype=mtype, as_scitype=scitype).get(fixture_index)

    # todo: possibly remove this once all checks are defined
    check_dict = get_check_dict()
    check_is_defined = (mtype, scitype) in check_dict.keys()

    # check fixtures that exist against checks that exist, when full metadata is queried
    if fixture is not None and check_is_defined:
        check_result = check_is_scitype(fixture, scitype, return_metadata=True)
        if not check_result[0]:
            msg = (
                f"check_is_scitype returns False on scitype {scitype}, mtype {mtype} "
                f"fixture {fixture_index}, message: "
            )
            msg = msg + str(check_result[1])
        assert check_result[0], msg
        assert check_result[2]["mtype"] == mtype

    # check fixtures that exist against checks that exist, when no metadata is queried
    if fixture is not None and check_is_defined:
        check_result = check_is_scitype(fixture, scitype, return_metadata=[])
        if not check_result[0]:
            msg = (
                f"check_is_scitype returns False on scitype {scitype}, mtype {mtype} "
                f"fixture {fixture_index}, message: "
            )
            msg = msg + str(check_result[1])
        assert check_result[0], msg
        assert check_result[2]["mtype"] == mtype


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes"),
    reason="Test only if skpro.datatypes has been changed",
)
def test_check_metadata_inference(scitype, mtype, fixture_index):
    """Tests that check_is_mtype correctly infers metadata of examples.

    Parameters
    ----------
    scitype : str - scitype of fixture
    mtype : str - mtype of fixture
    fixture_index : int - index of fixture tuple with that scitype and mtype

    Raises
    ------
    RuntimeError if scitype is not defined or has no mtypes or examples
    AssertionError if example metadata is not correctly inferred
    error if check itself raises an error
    """
    # retrieve fixture for checking
    fixture, lossy, expected_metadata = get_examples(
        mtype=mtype, as_scitype=scitype, return_metadata=True
    ).get(fixture_index)

    # todo: possibly remove this once all checks are defined
    check_dict = get_check_dict()
    check_is_defined = (mtype, scitype) in check_dict.keys()
    # if the examples have no metadata to them, don't test
    metadata_provided = expected_metadata is not None

    # metadata keys to ignore
    # is_equal_index is not fully supported yet in inference
    EXCLUDE_KEYS = ["is_equal_index"]

    # metadata keys to ignore if mtype is lossy
    EXCLUDE_IF_LOSSY = [
        "feature_names",  # lossy mtypes do not have feature names
    ]

    # if mtype is in the list, add mtype specific keys to exclude
    if lossy:
        EXCLUDE_KEYS += EXCLUDE_IF_LOSSY

    if metadata_provided:
        expected_metadata = expected_metadata.copy()
        subset_keys = set(expected_metadata.keys()).difference(EXCLUDE_KEYS)
        expected_metadata = {key: expected_metadata[key] for key in subset_keys}

    # check fixtures that exist against checks that exist, full metadata query
    if fixture is not None and check_is_defined and metadata_provided:
        check_result = check_is_mtype(
            fixture, mtype, scitype, return_metadata=True, msg_return_dict="list"
        )
        metadata = check_result[2]

        # remove mtype & scitype key if exists, since comparison is on scitype level
        if "mtype" in metadata:
            del metadata["mtype"]
        if "scitype" in metadata:
            del metadata["scitype"]

        # remove keys that are not checked
        for key in EXCLUDE_KEYS:
            if key in metadata:
                del metadata[key]

        # currently we do not check this field in metadata inference

        msg = (
            f"check_is_mtype returns wrong metadata on scitype {scitype}, "
            f"mtype {mtype}, fixture {fixture_index}, when querying full metadata. "
            f"returned: {metadata}; expected: {expected_metadata}"
        )

        for k, v in metadata.items():
            if v != "NA":
                assert v == expected_metadata[k], msg

    # check fixtures that exist against checks that exist
    if fixture is not None and check_is_defined and metadata_provided:
        for metadata_key in subset_keys:
            check_result = check_is_mtype(
                fixture,
                mtype,
                scitype,
                return_metadata=[metadata_key],
                msg_return_dict="list",
            )
            metadata = check_result[2]

            # remove mtype & scitype key if exists, since comparison is on scitype level
            if "mtype" in metadata:
                del metadata["mtype"]
            if "scitype" in metadata:
                del metadata["scitype"]

            msg = (
                f"check_is_mtype returns wrong metadata on scitype {scitype}, "
                f"mtype {mtype}, fixture {fixture_index}, when querying "
                f"metadata for metadata key {metadata_key}. "
                f"returned: {metadata[metadata_key]}; "
                f"expected: {expected_metadata[metadata_key]}"
            )

            if metadata[metadata_key] != "NA":
                assert metadata[metadata_key] == expected_metadata[metadata_key], msg


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes"),
    reason="Test only if skpro.datatypes has been changed",
)
def test_check_negative(scitype, mtype):
    """Tests that check_is_mtype correctly identifies wrong mtypes of examples.

    Parameters
    ----------
    scitype : str - scitype of fixture
    mtype : str - mtype of fixture

    Raises
    ------
    RuntimeError if scitype is not defined or has no mtypes or examples
    AssertionError if a examples are correctly identified as incompatible
    error if check itself raises an error
    """
    # if the scitype is ambiguous, we can't assume that other mtypes are negative
    if scitype in SCITYPES_AMBIGUOUS_MTYPE:
        return None

    mtypes = scitype_to_mtype(scitype)
    fixtures = dict()

    for other_mtype in mtypes:
        fixtures[other_mtype] = get_examples(mtype=other_mtype, as_scitype=scitype)

    n_fixtures = np.max([len(fixtures[mtype]) for mtype in mtypes])

    if n_fixtures == 0:
        raise RuntimeError("no fixtures defined for scitype " + scitype)

    for i in range(n_fixtures):
        # if mtype is not ambiguous, other mtypes are negative examples
        for wrong_mtype in list(set(mtypes).difference({mtype})):
            # retrieve fixture for checking
            fixture_wrong_type = fixtures[wrong_mtype].get(i)

            # todo: possibly remove this once all checks are defined
            check_dict = get_check_dict()
            check_is_defined = (mtype, scitype) in check_dict.keys()

            # check fixtures that exist against checks that exist
            if fixture_wrong_type is not None and check_is_defined:
                assert not check_is_mtype(
                    fixture_wrong_type,
                    mtype,
                    scitype,
                    msg_return_dict="list",
                ), (
                    f"check_is_mtype {mtype} returns True "
                    f"on {wrong_mtype} fixture {i}"
                )

            # check fixtures that exist against checks that exist
            if fixture_wrong_type is not None and check_is_defined:
                result = check_is_mtype(
                    fixture_wrong_type,
                    mtype,
                    scitype,
                    return_metadata=[],
                    msg_return_dict="list",
                )[0]
                assert not result, (
                    f"check_is_mtype {mtype} returns True "
                    f"on {wrong_mtype} fixture {i}"
                )


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes"),
    reason="Test only if skpro.datatypes has been changed",
)
def test_mtype_infer(scitype, mtype, fixture_index):
    """Tests that mtype correctly infers the mtype of examples.

    Parameters
    ----------
    scitype : str - scitype of fixture
    mtype : str - mtype of fixture
    fixture_index : int - index of fixture tuple with that scitype and mtype

    Raises
    ------
    RuntimeError if scitype is not defined or has no mtypes or examples
    AssertionError if mtype of examples is not correctly identified
    error if check itself raises an error
    """
    # if mtypes are ambiguous, then this test should be skipped
    if scitype in SCITYPES_AMBIGUOUS_MTYPE:
        return None

    # retrieve fixture for checking
    fixture = get_examples(mtype=mtype, as_scitype=scitype).get(fixture_index)

    # todo: possibly remove this once all checks are defined
    check_dict = get_check_dict()
    check_is_defined = (mtype, scitype) in check_dict.keys()

    # check fixtures that exist against checks that exist
    if fixture is not None and check_is_defined:
        assert mtype == infer_mtype(
            fixture, as_scitype=scitype, exclude_mtypes=[]
        ), f"mtype {mtype} not correctly identified for fixture {fixture_index}"

    # check indirect mtype inference via check_is_scitype
    if fixture is not None and check_is_defined:
        scitype_res = check_is_scitype(
            fixture, scitype=scitype, exclude_mtypes=[], return_metadata=[]
        )
        inferred_mtype = scitype_res[2]["mtype"]
        assert (
            mtype == inferred_mtype
        ), f"mtype {mtype} not correctly identified for fixture {fixture_index}"


# exclude these scitypes in inference of scitype test
#  would lead to ambiguous results
SKIP_SCITYPES = ["Alignment", "Table", "Proba"]
SCITYPES_FOR_INFER_TEST = list(set(SCITYPE_LIST).difference(SKIP_SCITYPES))


@pytest.mark.skipif(
    not run_test_module_changed("skpro.datatypes"),
    reason="Test only if skpro.datatypes has been changed",
)
def test_scitype_infer(scitype, mtype, fixture_index):
    """Tests that scitype correctly infers the mtype of examples.

    Parameters
    ----------
    scitype : str - scitype of fixture
    mtype : str - mtype of fixture
    fixture_index : int - index of fixture tuple with that scitype and mtype

    Raises
    ------
    RuntimeError if scitype is not defined or has no mtypes or examples
    AssertionError if mtype of examples is not correctly identified
    error if check itself raises an error
    """
    # if mtypes are ambiguous, then this test should be skipped
    if scitype in SKIP_SCITYPES or mtype in AMBIGUOUS_MTYPES:
        return None

    # retrieve fixture for checking
    fixture = get_examples(mtype=mtype, as_scitype=scitype).get(fixture_index)

    # todo: possibly remove this once all checks are defined
    check_dict = get_check_dict()
    check_is_defined = (mtype, scitype) in check_dict.keys()

    # check fixtures that exist against checks that exist
    if fixture is not None and check_is_defined:
        assert scitype == infer_scitype(
            fixture, candidate_scitypes=SCITYPES_FOR_INFER_TEST
        ), f"scitype {scitype} not correctly identified for fixture {fixture_index}"



"""Concrete performance metrics for probabilistic supervised regression."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
# adapted from sktime

import numpy as np
import pandas as pd

from skpro.metrics.base import BaseDistrMetric, BaseProbaMetric


class PinballLoss(BaseProbaMetric):
    r"""Pinball loss aka quantile loss for quantile/interval predictions.

    Can be used for both quantile and interval predictions.

    For a quantile prediction :math:`\widehat{y}
    at quantile point :math:`\alpha`,
    and a ground truth value :math:`y`, the pinball loss is defined as
    :math:`L_\alpha(y, \widehat{y}) := (y - \widehat{y}) \cdot (\alpha - H(y - \widehat{y}))`,
    where :math:`H` is the Heaviside step function defined as
    :math:`H(x) = 1` if :math:`x \ge 0` and :math:`H(x) = 0` otherwise.

    For a symmetric prediction interval :math:`I = [\widehat{y}_{\alpha}, \widehat{y}_{1 - \alpha}]`,
    the pinball loss is defined as
    :math:`L_\alpha(y, I) := L_\alpha(y, \widehat{y}_{\alpha}) + L_{1 - \alpha}(y, \widehat{y}_{1 - \alpha})`,
    or, in terms of coverage :math:`c = 1 - 2\alpha`, as
    :math:`L_c(y, I) := L_{1/2 - c/2}(y, a) + L_{1/2 + c/2}(y, b)`,
    if we write :math:`I = [a, b]`.

    * ``evaluate`` computes the average test sample loss.
    * ``evaluate_by_index`` produces the loss sample by test data point.
    * ``multivariate`` controls averaging over variables.
    * ``score_average`` controls averaging over quantiles/intervals.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    score_average : bool, optional, default = True
        specifies whether scores for each quantile should be averaged.

        * If True, metric/loss is averaged over all quantiles present in ``y_pred``.
        * If False, metric/loss is not averaged over quantiles.

    alpha (optional) : float, list of float, or 1D array-like, default=None
        quantiles to evaluate metric at.
        Can be specified if no explicit quantiles are present in the direct use of
        the metric, for instance in benchmarking via ``evaluate``, or tuning
        via ``ForecastingGridSearchCV``.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> from skpro.metrics import PinballLoss
    >>> y_true = pd.Series([3, -0.5, 2, 7, 2])
    >>> y_pred = pd.DataFrame({
    ...     ('Quantiles', 0.05): [1.25, 0, 1, 4, 0.625],
    ...     ('Quantiles', 0.5): [2.5, 0, 2, 8, 1.25],
    ...     ('Quantiles', 0.95): [3.75, 0, 3, 12, 1.875],
    ... })
    >>> pl = PinballLoss()
    >>> pl(y_true, y_pred)  # doctest: +SKIP
    np.float64(0.1791666666666667)
    >>> pl = PinballLoss(score_average=False)
    >>> pl(y_true, y_pred).to_numpy()
    array([0.16625, 0.275  , 0.09625])
    >>> y_true = pd.DataFrame({
    ...     "Quantiles1": [3, -0.5, 2, 7, 2],
    ...     "Quantiles2": [4, 0.5, 3, 8, 3],
    ... })
    >>> y_pred = pd.DataFrame({
    ...     ('Quantiles1', 0.05): [1.5, -1, 1, 4, 0.65],
    ...     ('Quantiles1', 0.5): [2.5, 0, 2, 8, 1.25],
    ...     ('Quantiles1', 0.95): [3.5, 4, 3, 12, 1.85],
    ...     ('Quantiles2', 0.05): [2.5, 0, 2, 8, 1.25],
    ...     ('Quantiles2', 0.5): [5.0, 1, 4, 16, 2.5],
    ...     ('Quantiles2', 0.95): [7.5, 2, 6, 24, 3.75],
    ... })
    >>> pl = PinballLoss(multioutput='raw_values')
    >>> pl(y_true, y_pred).to_numpy()
    array([0.16233333, 0.465     ])
    >>> pl = PinballLoss(multioutput=np.array([0.3, 0.7]))
    >>> pl(y_true, y_pred)  # doctest: +SKIP
    np.float64(0.3742000000000001)
    """  # noqa: E501

    _tags = {
        "scitype:y_pred": "pred_quantiles",
        "lower_is_better": True,
    }

    def __init__(self, multioutput="uniform_average", score_average=True, alpha=None):
        self.score_average = score_average
        self.alpha = alpha
        self._alpha = self._check_alpha(alpha)
        self.metric_args = {"alpha": self._alpha}
        super().__init__(multioutput=multioutput, score_average=score_average)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        """Logic for finding the metric evaluated at each index.

        y_true : pd.Series, pd.DataFrame or np.array of shape (fh,) or \
            (fh, n_outputs) where fh is the forecasting horizon
            Ground truth (correct) target value`s.

        y_pred : pd.Series, pd.DataFrame or np.array of shape (fh,) or  \
            (fh, n_outputs)  where fh is the forecasting horizon
            Forecasted values.

        multioutput : string "uniform_average" or "raw_values"
            Determines how multioutput results will be treated.
        """
        alpha = self._alpha
        y_pred_alphas = self._get_alpha_from(y_pred)
        if alpha is None:
            alphas = y_pred_alphas
        else:
            # if alpha was provided, check whether  they are predicted
            #   if not all alpha are observed, raise a ValueError
            if not np.isin(alpha, y_pred_alphas).all():
                # todo: make error msg more informative
                #   which alphas are missing
                msg = "not all quantile values in alpha are available in y_pred"
                raise ValueError(msg)
            else:
                alphas = alpha

        alphas = self._check_alpha(alphas)

        alpha_preds = y_pred.iloc[
            :, [x in alphas for x in y_pred.columns.get_level_values(1)]
        ]

        alpha_preds_np = alpha_preds.to_numpy()
        alpha_mat = np.repeat(
            (alpha_preds.columns.get_level_values(1).to_numpy().reshape(1, -1)),
            repeats=y_true.shape[0],
            axis=0,
        )

        y_true_np = np.repeat(y_true, axis=1, repeats=len(alphas))
        diff = y_true_np - alpha_preds_np
        sign = (diff >= 0).astype(diff.dtype)
        loss = alpha_mat * sign * diff - (1 - alpha_mat) * (1 - sign) * diff

        out_df = pd.DataFrame(loss, columns=alpha_preds.columns)

        return out_df

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Retrieve test parameters."""
        params1 = {}
        params2 = {"alpha": [0.1, 0.5, 0.9]}
        return [params1, params2]


class EmpiricalCoverage(BaseProbaMetric):
    r"""Empirical coverage percentage for interval predictions.

    Applies to interval predictions.

    Should be used together with ``ConstraintViolation`` if reported.

    Up to a constant, ``PinballLoss`` is a weighted sum of ``ConstraintViolation`` and
    ``EmpiricalCoverage``.

    For an interval prediction :math:`I = [a, b]` and a ground truth value :math:`y`,
    the empirical coverage loss is defined as

    :math:`L(y, I) := 1, \text{if } y \in I, 0 \text{ otherwise}`.

    When averaged over test samples, variables, or coverages, the average
    is the same as the empirical coverage percentage, i.e., the percentage of
    predictions that contain the true value, among the values averaged over.

    * ``evaluate`` computes the average test sample loss.
    * ``evaluate_by_index`` produces the loss sample by test data point.
    * ``multivariate`` controls averaging over variables.
    * ``score_average`` controls averaging over quantiles/intervals.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    score_average : bool, optional, default = True
        specifies whether scores for each coverage value should be averaged.

        * If True, metric/loss is averaged over all coverages present in ``y_pred``.
        * If False, metric/loss is not averaged over coverages.

    coverage (optional) : float, list of float, or 1D array-like, default=None
        nominal coverage to evaluate metric at.
        Can be specified if no explicit coverages are present in the direct use of
        the metric, for instance in benchmarking via ``evaluate``, or tuning
        via ``ForecastingGridSearchCV``.
    """

    _tags = {
        "scitype:y_pred": "pred_interval",
        "lower_is_better": False,
    }

    def __init__(
        self, multioutput="uniform_average", score_average=True, coverage=None
    ):
        self.score_average = score_average
        self.multioutput = multioutput
        self.coverage = coverage
        self._coverage = self._check_coverage(coverage)
        self.metric_args = {"coverage": self._coverage}
        super().__init__(score_average=score_average, multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        """Logic for finding the metric evaluated at each index.

        y_true : pd.Series, pd.DataFrame or np.array of shape (fh,) or \
            (fh, n_outputs) where fh is the forecasting horizon
            Ground truth (correct) target values.

        y_pred : pd.Series, pd.DataFrame or np.array of shape (fh,) or  \
            (fh, n_outputs)  where fh is the forecasting horizon
            Forecasted values.

        multioutput : string "uniform_average" or "raw_values" determines how \
            multioutput results will be treated.
        """
        lower = y_pred.iloc[:, y_pred.columns.get_level_values(2) == "lower"].to_numpy()
        upper = y_pred.iloc[:, y_pred.columns.get_level_values(2) == "upper"].to_numpy()

        if not isinstance(y_true, np.ndarray):
            y_true_np = y_true.to_numpy()
        else:
            y_true_np = y_true
        if y_true_np.ndim == 1:
            y_true_np = y_true.reshape(-1, 1)

        scores = np.unique(np.round(y_pred.columns.get_level_values(1), 7))
        no_scores = len(scores)
        vars = np.unique(y_pred.columns.get_level_values(0))

        y_true_np = np.tile(y_true_np, no_scores)

        truth_array = (y_true_np > lower).astype(int) * (y_true_np < upper).astype(int)

        out_df = pd.DataFrame(
            truth_array, columns=pd.MultiIndex.from_product([vars, scores])
        )

        return out_df

    @classmethod
    def get_test_params(self):
        """Retrieve test parameters."""
        params1 = {}
        params2 = {"coverage": 0.5}
        return [params1, params2]


class IntervalWidth(BaseProbaMetric):
    """Interval width for interval predictions, sometimes also known as sharpness.

    Applies to interval predictions.

    Also known as "sharpness".

    For an interval prediction :math:`I = [a, b]` and a ground truth value :math:`y`,
    the interval width is defined as

    :math:`W(y, I) := b - a`.

    * ``evaluate`` computes the average test sample loss.
    * ``evaluate_by_index`` produces the loss sample by test data point.
    * ``multivariate`` controls averaging over variables.
    * ``score_average`` controls averaging over quantiles/intervals.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    score_average : bool, optional, default = True
        specifies whether scores for each coverage value should be averaged.

        * If True, metric/loss is averaged over all coverages present in ``y_pred``.
        * If False, metric/loss is not averaged over coverages.

    coverage (optional) : float, list of float, or 1D array-like, default=None
        nominal coverage to evaluate metric at.
        Can be specified if no explicit coverages are present in the direct use of
        the metric, for instance in benchmarking via ``evaluate``, or tuning
        via ``ForecastingGridSearchCV``.
    """

    _tags = {
        "scitype:y_pred": "pred_interval",
        "lower_is_better": True,
    }

    def __init__(
        self, multioutput="uniform_average", score_average=True, coverage=None
    ):
        self.score_average = score_average
        self.multioutput = multioutput
        self.coverage = coverage
        self._coverage = self._check_coverage(coverage)
        self.metric_args = {"coverage": self._coverage}
        super().__init__(score_average=score_average, multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        """Logic for finding the metric evaluated at each index.

        y_true : pd.Series, pd.DataFrame or np.array of shape (fh,) or \
            (fh, n_outputs) where fh is the forecasting horizon
            Ground truth (correct) target values.

        y_pred : pd.Series, pd.DataFrame or np.array of shape (fh,) or  \
            (fh, n_outputs)  where fh is the forecasting horizon
            Forecasted values.

        multioutput : string "uniform_average" or "raw_values" determines how \
            multioutput results will be treated.
        """
        lower = y_pred.iloc[:, y_pred.columns.get_level_values(2) == "lower"].to_numpy()
        upper = y_pred.iloc[:, y_pred.columns.get_level_values(2) == "upper"].to_numpy()

        if not isinstance(y_true, np.ndarray):
            y_true_np = y_true.to_numpy()
        else:
            y_true_np = y_true
        if y_true_np.ndim == 1:
            y_true_np = y_true.reshape(-1, 1)

        scores = np.unique(np.round(y_pred.columns.get_level_values(1), 7))
        vars = np.unique(y_pred.columns.get_level_values(0))

        metric_array = upper - lower

        out_df = pd.DataFrame(
            metric_array, columns=pd.MultiIndex.from_product([vars, scores])
        )

        return out_df

    @classmethod
    def get_test_params(self):
        """Retrieve test parameters."""
        params1 = {}
        params2 = {"coverage": 0.5}
        return [params1, params2]


class ConstraintViolation(BaseProbaMetric):
    r"""Average absolute constraint violations for interval predictions.

    Applies to interval predictions.

    Should be used together with ``EmpiricalCoverage`` if reported.

    Up to a constant, ``PinballLoss`` is a weighted sum of ``ConstraintViolation`` and
    ``EmpiricalCoverage``.

    For an interval prediction :math:`I = [a, b]` and a ground truth value :math:`y`,
    the constraint violation loss is defined as

    .. math::

        L(y, I) :=
        \begin{cases}
        a - y, & \text{if } y < a \\
        y - b, & \text{if } y > b \\
        0, & \text{otherwise}
        \end{cases}

    * ``evaluate`` computes the average test sample loss.
    * ``evaluate_by_index`` produces the loss sample by test data point.
    * ``multivariate`` controls averaging over variables.
    * ``score_average`` controls averaging over quantiles/intervals.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    score_average : bool, optional, default = True
        specifies whether scores for each coverage value should be averaged.

        * If True, metric/loss is averaged over all coverages present in ``y_pred``.
        * If False, metric/loss is not averaged over coverages.

    coverage (optional) : float, list of float, or 1D array-like, default=None
        nominal coverage to evaluate metric at.
        Can be specified if no explicit coverages are present in the direct use of
        the metric, for instance in benchmarking via ``evaluate``, or tuning
        via ``ForecastingGridSearchCV``.
    """

    _tags = {
        "scitype:y_pred": "pred_interval",
        "lower_is_better": True,
    }

    def __init__(
        self, multioutput="uniform_average", score_average=True, coverage=None
    ):
        self.score_average = score_average
        self.multioutput = multioutput
        self.coverage = coverage
        self._coverage = self._check_coverage(coverage)
        self.metric_args = {"coverage": self._coverage}
        super().__init__(score_average=score_average, multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        """Logic for finding the metric evaluated at each index.

        y_true : pd.Series, pd.DataFrame or np.array of shape (fh,) or \
            (fh, n_outputs) where fh is the forecasting horizon
            Ground truth (correct) target values.

        y_pred : pd.Series, pd.DataFrame or np.array of shape (fh,) or  \
            (fh, n_outputs)  where fh is the forecasting horizon
            Forecasted values.

        multioutput : string "uniform_average" or "raw_values" determines how \
            multioutput results will be treated.
        """
        lower = y_pred.iloc[:, y_pred.columns.get_level_values(2) == "lower"].to_numpy()
        upper = y_pred.iloc[:, y_pred.columns.get_level_values(2) == "upper"].to_numpy()

        if not isinstance(y_true, np.ndarray):
            y_true_np = y_true.to_numpy()
        else:
            y_true_np = y_true

        if y_true_np.ndim == 1:
            y_true_np = y_true.reshape(-1, 1)

        scores = np.unique(np.round(y_pred.columns.get_level_values(1), 7))
        no_scores = len(scores)
        vars = np.unique(y_pred.columns.get_level_values(0))

        y_true_np = np.tile(y_true_np, no_scores)

        int_distance = ((y_true_np < lower).astype(int) * abs(lower - y_true_np)) + (
            (y_true_np > upper).astype(int) * abs(y_true_np - upper)
        )

        out_df = pd.DataFrame(
            int_distance, columns=pd.MultiIndex.from_product([vars, scores])
        )

        return out_df

    @classmethod
    def get_test_params(self):
        """Retrieve test parameters."""
        params1 = {}
        params2 = {"coverage": 0.5}
        return [params1, params2]


class LogLoss(BaseDistrMetric):
    r"""Logarithmic loss for distributional predictions.

    For a predictive distribution :math:`d` with pdf :math:`p_d`
    and a ground truth value :math:`y`, the logarithmic loss is
    defined as :math:`L(y, d) := -\log p_d(y)`.

    * ``evaluate`` computes the average test sample loss.
    * ``evaluate_by_index`` produces the loss sample by test data point.
    * ``multivariate`` controls averaging over variables.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    multivariate : bool, optional, default=False

        * if True, behaves as multivariate log-loss:
          the log-loss is computed for entire row, results one score per row
        * if False, is univariate log-loss:
          the log-loss is computed per variable marginal, results in many scores per row
    """

    def __init__(self, multioutput="uniform_average", multivariate=False):
        self.multivariate = multivariate
        super().__init__(multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        res = -y_pred.log_pdf(y_true)
        # replace this by multivariate log_pdf once distr implements
        # i.e., pass multivariate on to log_pdf
        if self.multivariate:
            return pd.DataFrame(res.mean(axis=1), columns=["density"])
        else:
            return res

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Retrieve test parameters."""
        params1 = {}
        params2 = {"multivariate": True}
        return [params1, params2]


class LinearizedLogLoss(BaseDistrMetric):
    r"""Lineararized logarithmic loss for distributional predictions.

    For a predictive distribution :math:`d` with pdf :math:`p_d`
    and a ground truth value :math:`y`, the linearized logarithmic loss is
    defined as :math:`L(y, d) := -\log p_d(y)` if :math:`p_d(y) \geq r`,
    and :math:`L(y, d) := -\log p_d(r) + 1 - \frac{1}{r} p_d(r)` otherwise,
    where :math:`r` is the range of linearization parameter, `range` below.

    * ``evaluate`` computes the average test sample loss.
    * ``evaluate_by_index`` produces the loss sample by test data point.
    * ``multivariate`` controls averaging over variables.

    Parameters
    ----------
    range : positive float, optional, default=1
        range of linearization, i.e., where to linearize the log-loss
        for values smaller than range, the log-loss is linearized

    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    multivariate : bool, optional, default=False

        * if True, behaves as multivariate squared loss:
          the score is computed for entire row, results one score per row
        * if False, is univariate squared loss:
          the score is computed per variable marginal, results in many scores per row
    """

    def __init__(self, range=1, multioutput="uniform_average", multivariate=False):
        self.range = range
        self.multivariate = multivariate
        super().__init__(multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        range = self.range

        pdf = y_pred.pdf(y_true)
        pdf_smaller_range = pdf < range
        pdf_greater_range = pdf >= range

        logloss = -y_pred.log_pdf(y_true)
        linear = (-1 / range) * pdf - np.log(range) + 1

        res = pdf_smaller_range * linear + pdf_greater_range * logloss

        # replace this by multivariate log_pdf once distr implements
        # i.e., pass multivariate on to log_pdf
        if self.multivariate:
            return pd.DataFrame(res.mean(axis=1), columns=["density"])
        else:
            return res

    @classmethod
    def get_test_params(self):
        """Test parameter settings."""
        params1 = {}
        params2 = {"range": 0.1}
        return [params1, params2]


class SquaredDistrLoss(BaseDistrMetric):
    r"""Squared loss for distributional predictions.

    Also known as:

    * continuous Brier loss
    * Gneiting loss
    * (mean) squared error/loss, i.e., confusingly named the same as the
      point prediction loss commonly known as the mean squared error

    For a predictive distribution :math:`d`
    and a ground truth value :math:`y`, the squared (distribution) loss is
    defined as :math:`L(y, d) := -2 p_d(y) + \|p_d\|^2`,
    where :math:`\|p_d\|^2` is the (function) L2-norm of :math:`p_d`.

    * ``evaluate`` computes the average test sample loss.
    * ``evaluate_by_index`` produces the loss sample by test data point.
    * ``multivariate`` controls averaging over variables.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    multivariate : bool, optional, default=False

        * if True, behaves as multivariate squared loss:
          the score is computed for entire row, results one score per row
        * if False, is univariate squared loss:
          the score is computed per variable marginal, results in many scores per row
    """

    def __init__(self, multioutput="uniform_average", multivariate=False):
        self.multivariate = multivariate
        super().__init__(multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        res = -2 * y_pred.log_pdf(y_true) + y_pred.pdfnorm(a=2)
        # replace this by multivariate log_pdf once distr implements
        # i.e., pass multivariate on to log_pdf
        if self.multivariate:
            return pd.DataFrame(res.mean(axis=1), columns=["density"])
        else:
            return res

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Retrieve test parameters."""
        params1 = {}
        params2 = {"multivariate": True}
        return [params1, params2]


class CRPS(BaseDistrMetric):
    r"""Continuous rank probability score for distributional predictions.

    Also known as:

    * integrated squared loss (ISL)
    * integrated Brier loss (IBL)
    * energy loss

    For a predictive distribution :math:`d` and a ground truth value :math:`y`,
    the CRPS is defined as
    :math:`L(y, d) := \mathbb{E}_{Y \sim d}|Y-y| - \frac{1}{2} \mathbb{E}_{X,Y \sim d}|X-Y|`.

    * ``evaluate`` computes the average test sample loss.
    * ``evaluate_by_index`` produces the loss sample by test data point.
    * ``multivariate`` controls averaging over variables.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    multivariate : bool, optional, default=False

        * if True, behaves as multivariate CRPS:
          the score is computed for entire row, results one score per row
        * if False, is univariate CRPS:
          the score is computed per variable marginal, results in many scores per row
    """  # noqa: E501

    def __init__(self, multioutput="uniform_average", multivariate=False):
        self.multivariate = multivariate
        super().__init__(multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        # CRPS(d, y) = E_X,Y as d [abs(Y-y) - 0.5 abs(X-Y)]
        return y_pred.energy(y_true) - y_pred.energy() / 2

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Retrieve test parameters."""
        params1 = {}
        params2 = {"multivariate": True}
        return [params1, params2]


class AUCalibration(BaseDistrMetric):
    r"""Area under the calibration curve for distributional predictions.

    Computes the unsigned area between the calibration curve and the diagonal.

    The calibration curve is the cumulative curve of the sample of
    predictive cumulative distribution functions evaluated at the true values.

    Mathematically, let :math:`d_1, \dots, d_N` be the predictive distributions,
    let :math:`y_1, \dots, y_N` be the true values, and let :math:`F_i` be the
    cumulative distribution function of :math:`d_i`.

    Define the calibration sample as :math:`c_i := F_i(y_i)`, for
    :math:`i = 1, \dots, N`. For perfect predictions, the sample of :math:`c_i` will be
    uniformly distributed on [0, 1], and i.i.d. from that uniform distribution.

    Let :math:`c_{(i)}` be the :math:`i`-th order statistic of the sample of
    :math:`c_i`, i.e., the :math:`i`-th smallest value in the sample.

    The (unsigned) area under the calibration curve - or, more precisely,
    between the diagonal and the calibration curve - is defined as

    .. math:: \frac{1}{N} \sum_{i=1}^N \left| c_{(i)} - \frac{i}{N} \right|.

    * ``evaluate`` returns the unsigned area between the calibration curve
      and the diagonal, i.e., the above quantity.
    * ``evaluate_by_index`` returns, for the :math:`i`-th test sample, the value
      :math:`\left| c_i - \frac{r_i}{N} \right|`,
      where :math:`r_i` is the rank of :math:`c_i`
      in the sample of :math:`c_i`. In case of ties, tied ranks are averaged.
    * ``multivariate`` controls averaging over variables.

    Parameters
    ----------
    multioutput : {'raw_values', 'uniform_average'} or array-like of shape \
            (n_outputs,), default='uniform_average'
        Defines whether and how to aggregate metric for across variables.

        * If 'uniform_average' (default), errors are mean-averaged across variables.
        * If array-like, errors are weighted averaged across variables,
          values as weights.
        * If 'raw_values', does not average errors across variables,
          columns are retained.

    multivariate : bool, optional, default=False

        * if True, behaves as multivariate metric (sum of scores):
          the metric is computed for entire row, results one score per row
        * if False, is univariate metric, per variable:
          the metric is computed per variable marginal, results in many scores per row
    """  # noqa: E501

    def __init__(self, multioutput="uniform_average", multivariate=False):
        self.multivariate = multivariate
        super().__init__(multioutput=multioutput)

    def _evaluate_by_index(self, y_true, y_pred, **kwargs):
        cdfs = y_pred.cdf(y_true)
        # using the average in case of ranks is fine
        # because the absolute sums in the metric average out
        cdfs_ranked = cdfs.rank(axis=0, method="average", pct=True)
        n = cdfs.shape[0]
        diagonal = np.arange(1, n + 1).reshape(-1, 1) / n

        res = (cdfs_ranked - diagonal).abs()

        if self.multivariate:
            return pd.DataFrame(res.mean(axis=1), columns=["score"])
        return res

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Retrieve test parameters."""
        params1 = {}
        params2 = {"multivariate": True}
        return [params1, params2]



# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Log-Normal probability distribution."""

__author__ = ["bhavikar04", "fkiraly"]

import numpy as np
import pandas as pd
from scipy.special import erf, erfinv

from skpro.distributions.base import BaseDistribution


class LogNormal(BaseDistribution):
    r"""Log-Normal distribution (skpro native).

    Parameterized by mean and standard deviation of the logarithm of the distribution,
    :math:`\mu` and :math:`\sigma`, respectively, such that the cdf

    .. math:: F(x) = \frac{1}{2} + \frac{1}{2} \text{erf}\left(\frac{\log(x) - \mu}{\sigma \sqrt{2}}\right)  # noqa E501

    Parameters
    ----------
    mu : float or array of float (1D or 2D)
        mean of the logarithm of the distribution, :math:`\mu` above
    sigma : float or array of float (1D or 2D), must be positive
        standard deviation the logarithm of the distribution, :math:`\sigma` above
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.lognormal import LogNormal

    >>> n = LogNormal(mu=[[0, 1], [2, 3], [4, 5]], sigma=1)
    """

    _tags = {
        "authors": ["bhavikar04", "fkiraly"],
        "capabilities:approx": ["energy", "pdfnorm"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        "distr:measuretype": "continuous",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, mu, sigma, index=None, columns=None):
        self.mu = mu
        self.sigma = sigma

        super().__init__(index=index, columns=columns)

    # commented out, seems incorrect
    # def energy(self, x=None):
    #     r"""Energy of self, w.r.t. self or a constant frame x.

    #     Let :math:`X, Y` be i.i.d. random variables with the distribution of `self`.

    #     If `x` is `None`, returns :math:`\mathbb{E}[|X-Y|]` (per row), "self-energy".
    #     If `x` is passed, returns :math:`\mathbb{E}[|X-x|]-0.5\mathbb{E}[|X-Y|]`
    #     (per row), "CRPS wrt x".

    #     Parameters
    #     ----------
    #     x : None or pd.DataFrame, optional, default=None
    #         if pd.DataFrame, must have same rows and columns as `self`

    #     Returns
    #     -------
    #     pd.DataFrame with same rows as `self`, single column `"energy"`
    #     each row contains one float, self-energy/energy as described above.
    #     """
    #     if x is None:
    #         return super().energy(x)
    #     # explicit formula for CRPS of log-normal cross-term
    #     # obtained by bhavikar04 via wolfram alpha
    #     else:
    #         d = self.loc[x.index, x.columns]
    #         mu_arr, sd_arr = d._mu, d._sigma
    #         c_arr = x * (2 * self.cdf(x) - 1)
    #         c_arr2 = -2 * np.exp((mu_arr + sd_arr**2) / 2)
    #         c_arr3 = self.cdf((np.log(x) - mu_arr - sd_arr**2) / sd_arr)
    #         c_arr3 = c_arr3 + self.cdf(sd_arr / mu_arr**0.5) - 1
    #         c_arr2 = c_arr2 * c_arr3
    #         c_arr = c_arr + c_arr2

    #         energy_arr = np.sum(c_arr, axis=1)
    #         energy = pd.DataFrame(energy_arr, index=self.index, columns=["energy"])
    #     return energy

    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        mean_arr = np.exp(mu + sigma**2 / 2)
        return mean_arr

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        sd_arr = np.exp(2 * mu + 2 * sigma**2) - np.exp(2 * mu + sigma**2)
        return sd_arr

    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        pdf_arr = np.exp(-0.5 * ((np.log(x) - mu) / sigma) ** 2)
        pdf_arr = pdf_arr / (x * sigma * np.sqrt(2 * np.pi))
        return pdf_arr

    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        lpdf_arr = -0.5 * ((np.log(x) - mu) / sigma) ** 2
        lpdf_arr = lpdf_arr - np.log(x * sigma * np.sqrt(2 * np.pi))
        return lpdf_arr

    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        cdf_arr = 0.5 + 0.5 * erf((np.log(x) - mu) / (sigma * np.sqrt(2)))
        return cdf_arr

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        mu = self._bc_params["mu"]
        sigma = self._bc_params["sigma"]

        icdf_arr = mu + sigma * np.sqrt(2) * erfinv(2 * p - 1)
        icdf_arr = np.exp(icdf_arr)
        return icdf_arr

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"mu": [[0, 1], [2, 3], [4, 5]], "sigma": 1}
        params2 = {
            "mu": 0,
            "sigma": 1,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"mu": -2, "sigma": 2}

        return [params1, params2, params3]



"""Utility functions for numpy/sklearn related matters."""

__authors__ = ["fkiraly"]


def flatten_to_1D_if_colvector(y):
    """Flattens a numpy array to 1D if it is a 2D column vector.

    Parameters
    ----------
    y : numpy array, 1D or 2D
        Array to flatten

    Returns
    -------
    y_flat : numpy array
        1D flattened array if y was 2D column vector, or 1D already
        otherwise, return y unchanged
    """
    if len(y.shape) == 2 and y.shape[1] == 1:
        y_flat = y.flatten()
    else:
        y_flat = y

    return y_flat



"""Extension template for probability distributions - simple pattern.

Purpose of this implementation template:
    quick implementation of new estimators following the template
    NOT a concrete class to import! This is NOT a base class or concrete class!
    This is to be used as a "fill-in" coding template.

How to use this implementation template to implement a new distribution:
- make a copy of the template in a suitable location, give it a descriptive name.
- work through all the "todo" comments below
- fill in code for mandatory methods, and optionally for optional methods
- do not write to reserved attributes: index, columns, head, tail, loc, iloc, at, iat,
  shape, ndim, _bc_params, _tags, _tags_dynamic, _config, _config_dynamic
- you can add more private methods, but do not override BaseEstimator's private methods
    an easy way to be safe is to prefix your methods with "_custom"
- change docstrings for functions and the file
- ensure interface compatibility by skpro.utils.estimator_checks.check_estimator
- once complete: use as a local library, or contribute to skpro via PR
- more details:
  https://www.sktime.net/en/stable/developer_guide/add_estimators.html

Mandatory methods to implement: at least one, better both of
    sampling        - sample(self, n_samples=None)
    ppf             - _ppf(self, p)

Optional methods to implement:
    mean            - _mean(self)
    variance        - _var(self)
    pdf             - _pdf(self, x)
    log_pdf         - _log_pdf(self, x)
    pmf             - _pmf(self, x)
    log_pmf         - _log_pmf(self, x)
    cdf             - _cdf(self, x)
    ppf             - _ppf(self, p)
    energy_self     - _energy_self(self)
    energy_x        - _energy_x(self, x)

Testing - required for test framework and check_estimator usage:
    get default parameters for test instance(s) - get_test_params()
"""
# todo: write an informative docstring for the file or module, remove the above
# todo: add an appropriate copyright notice for your estimator
#       estimators contributed to skpro should have the copyright notice at the top
#       estimators of your own do not need to have permissive or BSD-3 copyright

# todo: uncomment the following line, enter authors' GitHub IDs
# __author__ = [authorGitHubID, anotherAuthorGitHubID]

from skpro.distributions.base import BaseDistribution

# todo: add any necessary imports here - no soft dependency imports

# todo: for imports of skpro soft dependencies:
# make sure to fill in the "python_dependencies" tag with the package import name
# import soft dependencies only inside methods of the class, not at the top of the file


# todo: change class name and write docstring
class ClassName(BaseDistribution):
    """Custom probability distribution. todo: write docstring.

    todo: describe your custom probability distribution here

    Parameters
    ----------
    parama : float or np.ndarray
        descriptive explanation of parama
    paramb : float or np.ndarray, optional (default='default')
        descriptive explanation of paramb
    """

    # todo: fill out estimator tags here
    #  tags are inherited from parent class if they are not set
    # tags inherited from base are "safe defaults" which can usually be left as-is
    _tags = {
        # packaging info
        # --------------
        "authors": ["author1", "author2"],  # authors, GitHub handles
        "maintainers": ["maintainer1", "maintainer2"],  # maintainers, GitHub handles
        # author = significant contribution to code at some point
        # maintainer = algorithm maintainer role, "owner"
        # specify one or multiple authors and maintainers, only for skpro contribution
        # remove maintainer tag if maintained by skpro/sktim core team
        #
        "python_version": None,  # PEP 440 python version specifier to limit versions
        "python_dependencies": None,  # PEP 440 python dependencies specifier,
        # e.g., "numba>0.53", or a list, e.g., ["numba>0.53", "numpy>=1.19.0"]
        # delete if no python dependencies or version limitations
        #
        # estimator tags
        # --------------
        "distr:measuretype": "continuous",  # one of "discrete", "continuous", "mixed"
        # these tags should correspond to which methods are numerically exact
        # and which are approximations, e.g., using Monte Carlo
        "capabilities:approx": ["pdfnorm", "energy"],
        "capabilities:exact": ["mean", "var", "pdf", "log_pdf", "cdf", "ppf"],
        # leave the broadcast_init tag as-is, this tag exists for compatibility with
        # distributions deviating from assumptions on input parameters, e.g., Empirical
        "broadcast_init": "on",
    }

    # todo: fill init
    # params should be written to self and never changed
    # super call must not be removed, change class name
    # parameter checks can go after super call
    def __init__(self, param1, param2="param2default", index=None, columns=None):
        # all distributions must have index and columns arg with None defaults
        # this is to ensure pandas-like behaviour

        # todo: write any hyper-parameters and components to self
        self.param1 = param1
        self.param2 = param2

        # leave this as is
        super().__init__(index=index, columns=columns)

        # todo: optional, parameter checking logic (if applicable) should happen here
        # if writes derived values to self, should *not* overwrite self.parama etc
        # instead, write to self._parama, self._newparam (starting with _)

    # todo: implement as many of the following methods as possible
    # if not implemented, the base class will try to fill it in
    # from the other implemented methods
    # at least _ppf, or sample should be implemented for the distribution to be usable
    # if _ppf is implemented, sample does not need to be implemented (uses ppf sampling)

    # todo: consider implementing
    # if not implemented, uses Monte Carlo estimate via sample
    def _mean(self):
        """Return expected value of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            expected value of distribution (entry-wise)
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: consider implementing
    # if not implemented, uses Monte Carlo estimate via sample
    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            variance of the distribution (entry-wise)
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: consider implementing - only for continuous or mixed distributions
    # at least one of _pdf and _log_pdf should be implemented
    # if not implemented, returns exp of log_pdf
    def _pdf(self, x):
        """Probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pdf values at the given points
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: consider implementing - only for continuous or mixed distributions
    # at least one of _pdf and _log_pdf should be implemented
    # if not implemented, returns log of pdf
    def _log_pdf(self, x):
        """Logarithmic probability density function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pdf values at the given points
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: consider implementing - only for discrete or mixed distributions
    # at least one of _pmf and _log_pmf should be implemented
    # if not implemented, returns exp of log_pmf
    def _pmf(self, x):
        """Probability mass function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pmf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            pmf values at the given points
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: consider implementing - only for discrete or mixed distributions
    # at least one of _pmf and _log_pmf should be implemented
    # if not implemented, returns log of pmf
    def _log_pmf(self, x):
        """Logarithmic probability mass function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the pmf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            log pmf values at the given points
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: consider implementing
    # at least one of _ppf and sample must be implemented
    # if not implemented, uses Monte Carlo estimate based on sample
    def _cdf(self, x):
        """Cumulative distribution function.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to evaluate the cdf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            cdf values at the given points
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: consider implementing
    # at least one of _ppf and sample must be implemented
    # if not implemented, uses bisection method on cdf
    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf.

        Parameters
        ----------
        p : 2D np.ndarray, same shape as ``self``
            values to evaluate the ppf at

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            ppf values at the given points
        """
        pass

    # todo: consider implementing
    # if not implemented, uses Monte Carlo estimate via sample
    def _energy_self(self):
        r"""Energy of self, w.r.t. self.

        :math:`\mathbb{E}[|X-Y|]`, where :math:`X, Y` are i.i.d. copies of self.

        Private method, to be implemented by subclasses.

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: consider implementing
    # if not implemented, uses Monte Carlo estimate via sample
    def _energy_x(self, x):
        r"""Energy of self, w.r.t. a constant frame x.

        :math:`\mathbb{E}[|X-x|]`, where :math:`X` is a copy of self,
        and :math:`x` is a constant.

        Private method, to be implemented by subclasses.

        Parameters
        ----------
        x : 2D np.ndarray, same shape as ``self``
            values to compute energy w.r.t. to

        Returns
        -------
        2D np.ndarray, same shape as ``self``
            energy values w.r.t. the given points
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: consider implementing
    # at least one of _ppf and sample must be implemented
    # if not implemented, uses _ppf for sampling (inverse cdf on uniform)
    def sample(self, n_samples=None):
        """Sample from the distribution.

        Parameters
        ----------
        n_samples : int, optional, default = None
            number of samples to draw from the distribution

        Returns
        -------
        pd.DataFrame
            samples from the distribution

            * if ``n_samples`` is ``None``:
            returns a sample that contains a single sample from ``self``,
            in ``pd.DataFrame`` mtype format convention, with ``index`` and ``columns``
            as ``self``
            * if n_samples is ``int``:
            returns a ``pd.DataFrame`` that contains ``n_samples`` i.i.d.
            samples from ``self``, in ``pd-multiindex`` mtype format convention,
            with same ``columns`` as ``self``, and row ``MultiIndex`` that is product
            of ``RangeIndex(n_samples)`` and ``self.index``
        """
        param1 = self._bc_params["param1"]  # returns broadcast params to x.shape
        param2 = self._bc_params["param2"]  # returns broadcast params to x.shape

        res = "do_sth_with(" + param1 + param2 + ")"  # replace this by internal logic
        return res

    # todo: return default parameters, so that a test instance can be created
    #   required for automated unit and integration testing of estimator
    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """

        # todo: set the testing parameters for the estimators
        # Testing parameters can be dictionary or list of dictionaries
        #
        # this can, if required, use:
        #   class properties (e.g., inherited); parent class test case
        #   imported objects such as estimators from skpro or sklearn
        # important: all such imports should be *inside get_test_params*, not at the top
        #            since imports are used only at testing time
        #
        # A parameter dictionary must be returned *for all values* of parameter_set,
        #   i.e., "parameter_set not available" errors should never be raised.
        #
        # A good parameter set should primarily satisfy two criteria,
        #   1. Chosen set of parameters should have a low testing time,
        #      ideally in the magnitude of few seconds for the entire test suite.
        #       This is vital for the cases where default values result in
        #       "big" models which not only increases test time but also
        #       run into the risk of test workers crashing.
        #   2. There should be a minimum two such parameter sets with different
        #      sets of values to ensure a wide range of code coverage is provided.
        #
        # example 1: specify params as dictionary
        # any number of params can be specified
        # params = {"est": value0, "parama": value1, "paramb": value2}
        #
        # example 2: specify params as list of dictionary
        # note: Only first dictionary will be used by create_test_instance
        # params = [{"est": value1, "parama": value2},
        #           {"est": value3, "parama": value4}]
        #
        # example 3: parameter set depending on param_set value
        #   note: only needed if a separate parameter set is needed in tests
        # if parameter_set == "special_param_set":
        #     params = {"est": value1, "parama": value2}
        #     return params
        #
        # # "default" params
        # params = {"est": value3, "parama": value4}
        # return params



"""Bagging probabilistic regressors."""

__author__ = ["fkiraly"]
__all__ = ["BaggingRegressor"]

from math import ceil

import numpy as np
import pandas as pd

from skpro.distributions.mixture import Mixture
from skpro.regression.base import BaseProbaRegressor


class BaggingRegressor(BaseProbaRegressor):
    """Bagging ensemble of probabilistic regresesors.

    Fits ``n_estimators`` clones of an skpro regressor on
    datasets which are instance sub-samples and/or variable sub-samples.

    On ``predict_proba``, the mixture of probabilistic predictions is returned.

    The estimator allows to choose sample sizes for instances, variables,
    and whether sampling is with or without replacement.

    Direct generalization of ``sklearn``'s ``BaggingClassifier``
    to the probabilistic regrsesion task.

    Parameters
    ----------
    estimator : skpro regressor, descendant of BaseProbaRegressor
        regressor to use in the bagging estimator
    n_estimators : int, default=10
        number of estimators in the sample for bagging
    n_samples : int or float, default=1.0
        The number of instances drawn from ``X`` in ``fit`` to train each clone
        If int, then indicates number of instances precisely
        If float, interpreted as a fraction, and rounded by ``ceil``
    n_features : int or float, default=1.0
        The number of features/variables drawn from ``X`` in ``fit`` to train each clone
        If int, then indicates number of instances precisely
        If float, interpreted as a fraction, and rounded by ``ceil``
    bootstrap : boolean, default=True
        whether samples/instances are drawn with replacement (True) or not (False)
    bootstrap_features : boolean, default=False
        whether features/variables are drawn with replacement (True) or not (False)
    random_state : int, RandomState instance or None, optional (default=None)
        If int, ``random_state`` is the seed used by the random number generator;
        If ``RandomState`` instance, ``random_state`` is the random number generator;
        If None, the random number generator is the ``RandomState`` instance used
        by ``np.random``.

    Attributes
    ----------
    estimators_ : list of of skpro regressors
        clones of regressor in `estimator` fitted in the ensemble

    Examples
    --------
    >>> from skpro.regression.ensemble import BaggingRegressor
    >>> from skpro.regression.residual import ResidualDouble
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.model_selection import train_test_split
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)
    >>>
    >>> reg_mean = LinearRegression()
    >>> reg_proba = ResidualDouble(reg_mean)
    >>>
    >>> ens = BaggingRegressor(reg_proba, n_estimators=10)
    >>> ens.fit(X_train, y_train)
    BaggingRegressor(...)
    >>> y_pred = ens.predict_proba(X_test)
    """

    _tags = {"capability:missing": True}

    def __init__(
        self,
        estimator,
        n_estimators=10,
        n_samples=1.0,
        n_features=1.0,
        bootstrap=True,
        bootstrap_features=False,
        random_state=None,
    ):
        self.estimator = estimator
        self.n_estimators = n_estimators
        self.n_samples = n_samples
        self.n_features = n_features
        self.bootstrap = bootstrap
        self.bootstrap_features = bootstrap_features
        self.random_state = random_state

        super().__init__()

        tags_to_clone = ["capability:missing"]
        self.clone_tags(estimator, tags_to_clone)

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        estimator = self.estimator
        n_estimators = self.n_estimators
        n_samples = self.n_samples
        n_features = self.n_features
        bootstrap = self.bootstrap
        bootstrap_ft = self.bootstrap_features
        random_state = self.random_state
        np.random.seed(random_state)

        inst_ix = X.index
        col_ix = X.columns
        n = len(inst_ix)
        m = len(col_ix)

        if isinstance(n_samples, float):
            n_samples_ = ceil(n_samples * n)
        else:
            n_samples_ = n_samples

        if isinstance(n_features, float):
            n_features_ = ceil(n_features * m)
        else:
            n_features_ = n_features

        self.estimators_ = []
        self.cols_ = []

        for _i in range(n_estimators):
            esti = estimator.clone()
            row_iloc = pd.RangeIndex(n)
            row_ss = _random_ss_ix(row_iloc, size=n_samples_, replace=bootstrap)
            inst_ix_i = inst_ix[row_ss]
            col_ix_i = _random_ss_ix(col_ix, size=n_features_, replace=bootstrap_ft)

            # store column subset for use in predict
            self.cols_ += [col_ix_i]

            Xi = _subs_cols(X.loc[inst_ix_i], col_ix_i, reset_cols=bootstrap_ft)
            Xi = Xi.reset_index(drop=True)

            yi = y.loc[inst_ix_i].reset_index(drop=True)

            self.estimators_ += [esti.fit(Xi, yi)]

        return self

    def _predict_proba(self, X) -> np.ndarray:
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        reset_cols = self.bootstrap_features
        Xis = [_subs_cols(X, col_ix_i, reset_cols) for col_ix_i in self.cols_]

        y_probas = [est.predict_proba(Xi) for est, Xi in zip(self.estimators_, Xis)]

        y_proba = Mixture(y_probas)

        return y_proba

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.linear_model import LinearRegression

        from skpro.regression.residual import ResidualDouble

        regressor = ResidualDouble(LinearRegression())

        params1 = {"estimator": regressor}
        params2 = {
            "estimator": regressor,
            "n_samples": 0.5,
            "n_features": 0.5,
        }
        params3 = {
            "estimator": regressor,
            "n_samples": 7,
            "n_features": 2,
            "bootstrap": False,
            "bootstrap_features": True,
        }

        return [params1, params2, params3]


def _random_ss_ix(ix, size, replace=True):
    """Randomly uniformly sample indices from a list of indices."""
    a = range(len(ix))
    ixs = ix[np.random.choice(a, size=size, replace=replace)]
    return ixs


def _subs_cols(df, col_ix, reset_cols=False):
    """Subset columns of a DataFrame, with potential resetting of column index."""
    df_subset = df.loc[:, col_ix]
    if reset_cols:
        df_subset.columns = pd.RangeIndex(len(df_subset.columns))
    return df_subset



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Binomial probability distribution."""

__author__ = ["meraldoantonio"]

import pandas as pd
from scipy.stats import binom, rv_discrete

from skpro.distributions.adapters.scipy import _ScipyAdapter


class Binomial(_ScipyAdapter):
    r"""Binomial distribution.

    Most methods wrap ``scipy.stats.binom``.
    The Binomial distribution is parameterized by the number of trials :math:`n`
    and the probability of success :math:`p`,
    such that the probability mass function (PMF) is given by:

    .. math:: P(X = k) = \binom{n}{k} p^k (1-p)^{n-k}

    Parameters
    ----------
    n : int or array of int (1D or 2D), must be non-negative
    p : float or array of float (1D or 2D), must be in [0, 1]
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> from skpro.distributions.binomial import Binomial

    >>> d = Binomial(n=[[10, 10], [20, 30], [40, 50]], p=0.5)
    """

    _tags = {
        "capabilities:approx": ["pmf"],
        "capabilities:exact": ["mean", "var", "pmf", "log_pmf", "cdf", "ppf"],
        "distr:measuretype": "discrete",
        "distr:paramtype": "parametric",
        "broadcast_init": "on",
    }

    def __init__(self, n, p, index=None, columns=None):
        self.n = n
        self.p = p

        super().__init__(index=index, columns=columns)

    def _get_scipy_object(self) -> rv_discrete:
        return binom

    def _get_scipy_param(self):
        n = self._bc_params["n"]
        p = self._bc_params["p"]

        return [], {"n": n, "p": p}

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # array case examples
        params1 = {"n": [[10, 10], [20, 30], [40, 50]], "p": 0.5}
        params2 = {
            "n": 10,
            "p": 0.5,
            "index": pd.Index([1, 2, 5]),
            "columns": pd.Index(["a", "b"]),
        }
        # scalar case examples
        params3 = {"n": 15, "p": 0.7}

        return [params1, params2, params3]



"""Interface adapter to lifelines Aalen additive surival model."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["fkiraly"]

from skpro.survival.adapters.lifelines import _LifelinesAdapter
from skpro.survival.base import BaseSurvReg


class AalenAdditive(_LifelinesAdapter, BaseSurvReg):
    r"""Aalen additive hazards model, from lifelines.

    Direct interface to ``lifelines.fitters.AalenAdditiveFitter``,
    by ``CamDavidsonPilon``.

    This class fits the regression model:

    .. math::  h(t|x)  = b_0(t) + b_1(t) x_1 + ... + b_N(t) x_N

    that is, the hazard rate is a linear function of the covariates
    with time-varying coefficients.
    This implementation assumes non-time-varying covariates.

    Parameters
    ----------
    fit_intercept: bool, optional (default: True)
      If False, do not attach an intercept (column of ones) to the covariate matrix.
      The intercept, :math:`b_0(t)` acts as a baseline hazard.
    alpha: float, optional (default=0.05)
      the level in the confidence intervals around the estimated survival function,
      for computation of ``confidence_intervals_`` fitted parameter.
    coef_penalizer: float, optional (default: 0)
      Attach a L2 penalizer to the size of the coefficients during regression.
      This improves
      stability of the estimates and controls for high correlation between covariates.
      For example, this shrinks the magnitude of :math:`c_{i,t}`.
    smoothing_penalizer: float, optional (default: 0)
      Attach a L2 penalizer to difference between adjacent (over time) coefficients.
      For example, this shrinks the magnitude of :math:`c_{i,t} - c_{i,t+1}`.

    Attributes
    ----------
    cumulative_hazards_ : DataFrame
        The estimated cumulative hazard
    hazards_ : DataFrame
        The estimated hazards
    confidence_intervals_ : DataFrame
        The lower and upper confidence intervals for the cumulative hazard
    durations: array
        The durations provided
    """

    _tags = {"authors": ["CamDavidsonPilon", "rocreguant", "fkiraly"]}
    # CamDavidsonPilon, rocreguant credit for interfaced estimator

    def __init__(
        self,
        fit_intercept=True,
        alpha=0.05,
        coef_penalizer=0.0,
        smoothing_penalizer=0.0,
    ):
        self.fit_intercept = fit_intercept
        self.alpha = alpha
        self.coef_penalizer = coef_penalizer
        self.smoothing_penalizer = smoothing_penalizer

        super().__init__()

    def _get_lifelines_class(self):
        """Getter of the lifelines class to be used for the adapter."""
        from lifelines.fitters.aalen_additive_fitter import AalenAdditiveFitter

        return AalenAdditiveFitter

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}

        params2 = {
            "fit_intercept": False,
            "alpha": 0.1,
            "coef_penalizer": 0.1,
            "smoothing_penalizer": 0.1,
        }

        return [params1, params2]



"""Probabilistic predictiins by point prediction."""

__author__ = ["fkiraly"]
__all__ = ["DeltaPointRegressor"]

import pandas as pd
from sklearn import clone

from skpro.distributions.delta import Delta
from skpro.regression.base import BaseProbaRegressor
from skpro.utils.sklearn import prep_skl_df


class DeltaPointRegressor(BaseProbaRegressor):
    """Delta distribution prediction baseline regressor.

    This regressor turns an ``sklearn`` point prediction regressor into a probabilistic
    regressor by taking the point prediction as a predictive delta distribution.

    That is, on ``predict_proba``, a delta distribution with location
    identical to the regressor's ``predict`` is returned.

    Parameters
    ----------
    estimator : sklearn regressor
        regressor to use in the bootstrap

    Attributes
    ----------
    estimator_ : fitted sklearn regressor
        clones of regressor ``estimator``, fitted to the data

    Examples
    --------
    >>> from sklearn.linear_model import LinearRegression
    >>> from sklearn.datasets import load_diabetes
    >>> from sklearn.model_selection import train_test_split
    >>> from skpro.regression.delta import DeltaPointRegressor
    >>>
    >>> X, y = load_diabetes(return_X_y=True, as_frame=True)
    >>> X_train, X_test, y_train, y_test = train_test_split(X, y)
    >>>
    >>> reg_tabular = LinearRegression()
    >>>
    >>> reg_proba = DeltaPointRegressor(reg_tabular)
    >>> reg_proba.fit(X_train, y_train)
    DeltaPointRegressor(...)
    >>> y_pred = reg_proba.predict_proba(X_test)
    """

    _tags = {"authors": "fkiraly", "capability:missing": True}

    def __init__(self, estimator):
        self.estimator = estimator

        super().__init__()

        # todo: find the equivalent tag in sklearn for missing data handling
        # tags_to_clone = ["capability:missing"]
        # self.clone_tags(estimator, tags_to_clone)

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        self._cols = y.columns
        estimator = self.estimator

        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        # fit a clone of the estimator
        est_ = clone(estimator)
        self.estimator_ = est_.fit(X, y)
        return self

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        cols = self._cols

        # coerce X to pandas DataFrame with string column names
        X = prep_skl_df(X, copy_df=True)

        # predict point predictions
        y_pred = self.estimator_.predict(X)

        if not isinstance(y_pred, pd.DataFrame):
            y_pred = pd.DataFrame(y_pred, columns=cols, index=X.index)

        # create delta distribution
        y_proba = Delta(c=y_pred)
        return y_proba

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.ensemble import RandomForestRegressor
        from sklearn.linear_model import LinearRegression

        params1 = {"estimator": LinearRegression()}
        params2 = {"estimator": RandomForestRegressor()}

        return [params1, params2]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Doctest checks directed through pytest with conditional skipping."""

import importlib
import inspect
import pkgutil
from functools import lru_cache

from skpro.tests.test_all_estimators import ONLY_CHANGED_MODULES
from skpro.tests.test_switch import run_test_module_changed
from skpro.utils._doctest import run_doctest

EXCLUDE_MODULES_STARTING_WITH = ("all", "test")


def _all_functions(module_name):
    """Get all functions from a module, including submodules.

    Excludes:

    * modules starting with 'all' or 'test'.
    * if the flag ``ONLY_CHANGED_MODULES`` is set, modules that have not changed,
      compared to the ``main`` branch.

    Parameters
    ----------
    module_name : str
        Name of the module.

    Returns
    -------
    functions_list : list
        List of tuples (function_name, function_object).
    """
    res = _all_functions_cached(module_name, only_changed_modules=ONLY_CHANGED_MODULES)
    # copy the result to avoid modifying the cached result
    return res.copy()


@lru_cache
def _all_functions_cached(module_name, only_changed_modules=False):
    """Get all functions from a module, including submodules.

    Excludes:

    * modules starting with 'all' or 'test'.
    * if ``only_changed_modules`` is ``True``, modules that have not changed,
      compared to the ``main`` branch.

    Parameters
    ----------
    module_name : str
        Name of the module.
    only_changed_modules : bool, optional (default=False)
        If True, only functions from modules that have changed are returned.

    Returns
    -------
    functions_list : list
        List of tuples (function_name, function_object).
    """
    # Import the package
    package = importlib.import_module(module_name)

    # Initialize an empty list to hold all functions
    functions_list = []

    # Walk through the package's modules
    package_path = package.__path__[0]
    for _, modname, _ in pkgutil.walk_packages(
        path=[package_path], prefix=package.__name__ + "."
    ):
        # Skip modules starting with 'all' or 'test'
        if modname.split(".")[-1].startswith(EXCLUDE_MODULES_STARTING_WITH):
            continue

        # Skip modules that have not changed
        if only_changed_modules and not run_test_module_changed(modname):
            continue

        # Import the module
        module = importlib.import_module(modname)

        # Get all functions from the module
        for name, obj in inspect.getmembers(module, inspect.isfunction):
            # if function is imported from another module, skip it
            if obj.__module__ != module.__name__:
                continue
            # add the function to the list
            functions_list.append((name, obj))

    return functions_list


def pytest_generate_tests(metafunc):
    """Test parameterization routine for pytest.

    Fixtures parameterized
    ----------------------
    func : all functions from sktime, as returned by _all_functions
        if ONLY_CHANGED_MODULES is set, only functions from modules that have changed
    """
    # we assume all four arguments are present in the test below
    funcs_and_names = _all_functions("skpro")

    if len(funcs_and_names) > 0:
        names, funcs = zip(*funcs_and_names)

        metafunc.parametrize("func", funcs, ids=names)
    else:
        metafunc.parametrize("func", [])


def test_all_functions_doctest(func):
    """Run doctest for all functions in skpro."""
    run_doctest(func, name=f"function {func.__name__}")



# copyright: sktime developers, BSD-3-Clause License (see LICENSE file)
"""Implements adapter for scikit-survival models."""

__all__ = ["_SksurvAdapter"]
__author__ = ["fkiraly"]

import numpy as np
import pandas as pd

from skpro.distributions.empirical import Empirical
from skpro.survival.adapters._common import _get_fitted_params_default_safe
from skpro.utils.sklearn import prep_skl_df


class _SksurvAdapter:
    """Mixin adapter class for sksurv models."""

    _tags = {
        # packaging info
        # --------------
        "authors": ["fkiraly"],
        "python_dependencies": ["scikit-survival"],
        "python_dependencies_alias": {"scikit-survival": "sksurv"},
        "license_type": "copyleft",
        # capability tags
        # ---------------
        "X_inner_mtype": "pd_DataFrame_Table",
        "y_inner_mtype": "pd_DataFrame_Table",
        "C_inner_mtype": "pd_DataFrame_Table",
        "capability:multioutput": False,
    }

    # defines the name of the attribute containing the sksurv estimator
    _estimator_attr = "_estimator"

    def _get_sksurv_class(self):
        """Abstract method to get sksurv class.

        should import and return sksurv class
        """
        # from sksurv import SksurvClass
        #
        # return SksurvClass
        raise NotImplementedError("abstract method")

    def _get_sksurv_object(self):
        """Abstract method to initialize sksurv object.

        The default initializes result of _get_sksurv_class
        with self.get_params.
        """
        cls = self._get_sksurv_class()
        return cls(**self.get_params())

    def _init_sksurv_object(self):
        """Abstract method to initialize sksurv object and set to _estimator_attr.

        The default writes the return of _get_sksurv_object to
        the attribute of self with name _estimator_attr
        """
        cls = self._get_sksurv_object()
        setattr(self, self._estimator_attr, cls)
        return getattr(self, self._estimator_attr)

    def _fit(self, X, y, C=None):
        """Fit estimator training data.

        Parameters
        ----------
        X : pd.DataFrame
            Training features
        y: pd.Series
            Training labels
        C: pd.Series, optional (default=None)
            Censoring information for survival analysis.

        Returns
        -------
        self: reference to self
            Fitted estimator.
        """
        sksurv_est = self._init_sksurv_object()

        if C is None:
            C = pd.DataFrame(np.zeros(len(y)), index=y.index, columns=y.columns)

        # input conversion
        X = X.astype("float")  # sksurv insists on float dtype
        X = prep_skl_df(X)
        y_np = y.iloc[:, 0].values  # we know univariate due to tag
        C_np = C.iloc[:, 0].values
        C_np_bool = C_np == 0  # sksurv uses "delta" indicator, 0 = censored
        # this is the opposite of skpro ("censoring" indicator), where 1 = censored

        y_sksurv = list(zip(C_np_bool, y_np))
        y_sksurv = np.array(y_sksurv, dtype=[("delta", "?"), ("time", "<f8")])

        self._y_cols = y.columns  # remember column names for later

        # fit sksurv estimator
        sksurv_est.fit(X, y_sksurv)

        # write fitted params to self
        # some fitted parameters are properties and may raise exceptions
        # for example, AIC_ and AIC_partial_ of CoxPHFitter
        # to avoid this, we use a safe getter
        EXCEPTED_FITTED_PARAMS = ["n_features_in", "feature_names_in"]
        sksurv_fitted_params = _get_fitted_params_default_safe(sksurv_est)
        for k, v in sksurv_fitted_params.items():
            if k not in EXCEPTED_FITTED_PARAMS:
                setattr(self, f"{k}_", v)

        return self

    def _predict_proba(self, X):
        """Predict_proba method adapter.

        Parameters
        ----------
        X : pd.DataFrame
            Features to predict on.

        Returns
        -------
        skpro Empirical distribution
        """
        sksurv_est = getattr(self, self._estimator_attr)

        # input conversion
        X = X.astype("float")  # sksurv insists on float dtype
        X = prep_skl_df(X)

        # predict on X
        sksurv_survf = sksurv_est.predict_survival_function(X, return_array=True)

        times = sksurv_est.unique_times_[:-1]

        nt = len(times)
        mi = pd.MultiIndex.from_product([X.index, range(nt)]).swaplevel()

        times_val = np.repeat(times, repeats=len(X))
        times_df = pd.DataFrame(times_val, index=mi, columns=self._y_cols)

        weights = -np.diff(sksurv_survf, axis=1).flatten()
        weights_df = pd.Series(weights, index=mi)

        dist = Empirical(
            spl=times_df, weights=weights_df, index=X.index, columns=self._y_cols
        )

        return dist



"""Utility to determine scitype of estimator, based on base class type."""

__author__ = ["fkiraly"]

from inspect import isclass


def scitype(obj, force_single_scitype=True, coerce_to_list=False):
    """Determine scitype string of obj.

    Parameters
    ----------
    obj : class or object inheriting from sktime BaseObject
    force_single_scitype : bool, optional, default = True
        whether only a single scitype is returned
        if True, only the *first* scitype found will be returned
        order is determined by the order in BASE_CLASS_REGISTER
    coerce_to_list : bool, optional, default = False
        whether return should be coerced to list, even if only one scitype is identified

    Returns
    -------
    scitype : str, or list of str of sktime scitype strings from BASE_CLASS_REGISTER
        str, sktime scitype string, if exactly one scitype can be determined for obj
            or force_single_scitype is True, and if coerce_to_list is False
        list of str, of scitype strings, if more than one scitype are determined,
            or if coerce_to_list is True
        obj has scitype if it inherits from class in same row of BASE_CLASS_REGISTER

    Raises
    ------
    TypeError if no scitype can be determined for obj
    """
    # if object has tag, return tag
    if hasattr(obj, "get_tag"):
        if isclass(obj):
            tag_type = obj.get_class_tag("object_type", None)
        else:
            tag_type = obj.get_tag("object_type", None, raise_error=False)
        if tag_type is not None:
            if coerce_to_list and not isinstance(tag_type, list):
                scitypes = [tag_type]
            else:
                scitypes = tag_type
    else:
        scitypes = ["object"]

    if isinstance(scitypes, list) and len(scitypes) == 0:
        raise TypeError("Error, no scitype could be determined for obj")

    if isinstance(scitypes, list) and force_single_scitype:
        scitypes = [scitypes[0]]

    if isinstance(scitypes, list) and len(scitypes) == 1 and not coerce_to_list:
        return scitypes[0]

    return scitypes



"""Register of estimator and object tags.

Note for extenders: new tags should be entered in OBJECT_TAG_REGISTER.
No other place is necessary to add new tags.

This module exports the following:

---
OBJECT_TAG_REGISTER - list of tuples

each tuple corresponds to a tag, elements as follows:
    0 : string - name of the tag as used in the _tags dictionary
    1 : string - name of the scitype this tag applies to
                 must be in _base_classes.BASE_CLASS_SCITYPE_LIST
    2 : string - expected type of the tag value
        should be one of:
            "bool" - valid values are True/False
            "int" - valid values are all integers
            "str" - valid values are all strings
            "list" - valid values are all lists of arbitrary elements
            ("str", list_of_string) - any string in list_of_string is valid
            ("list", list_of_string) - any individual string and sub-list is valid
            ("list", "str") - any individual string or list of strings is valid
        validity can be checked by check_tag_is_valid (see below)
    3 : string - plain English description of the tag

---

OBJECT_TAG_TABLE - pd.DataFrame
    OBJECT_TAG_REGISTER in table form, as pd.DataFrame
        rows of OBJECT_TABLE correspond to elements in OBJECT_TAG_REGISTER

OBJECT_TAG_LIST - list of string
    elements are 0-th entries of OBJECT_TAG_REGISTER, in same order

---

check_tag_is_valid(tag_name, tag_value) - checks whether tag_value is valid for tag_name
"""
import pandas as pd

OBJECT_TAG_REGISTER = [
    # --------------------------
    # all objects and estimators
    # --------------------------
    (
        "reserved_params",
        "object",
        "list",
        "list of reserved parameter names",
    ),
    (
        "object_type",
        "object",
        "str",
        "type of object, e.g., 'regressor', 'transformer'",
    ),
    (
        "estimator_type",
        "estimator",
        "str",
        "type of estimator, e.g., 'regressor', 'transformer'",
    ),
    # packaging information
    # ---------------------
    (
        "maintainers",
        "object",
        ("list", "str"),
        "list of current maintainers of the object, each maintainer a GitHub handle",
    ),
    (
        "authors",
        "object",
        ("list", "str"),
        "list of authors of the object, each author a GitHub handle",
    ),
    (
        "python_version",
        "object",
        "str",
        "python version specifier (PEP 440) for estimator, or None = all versions ok",
    ),
    (
        "python_dependencies",
        "object",
        ("list", "str"),
        "python dependencies of estimator as str or list of str",
    ),
    (
        "python_dependencies_alias",
        "object",
        "dict",
        "should be provided if import name differs from package name, \
        key-value pairs are package name, import name",
    ),
    (
        "license_type",
        "object",
        "str",
        "license type for interfaced packages: 'copyleft', 'permissive', 'copyright'. \
        may be incorrect, NO LIABILITY assumed for this field",
    ),
    # ------------------
    # BaseProbaRegressor
    # ------------------
    (
        "capability:survival",
        "regressor_proba",
        "bool",
        "whether estimator can use censoring information, for survival analysis",
    ),
    (
        "capability:multioutput",
        "regressor_proba",
        "bool",
        "whether estimator supports multioutput regression",
    ),
    (
        "capability:missing",
        "regressor_proba",
        "bool",
        "whether estimator supports missing values",
    ),
    (
        "capability:update",
        "regressor_proba",
        "bool",
        "whether estimator supports online updates via update",
    ),
    (
        "X_inner_mtype",
        "regressor_proba",
        ("list", "str"),
        "which machine type(s) is the internal _fit/_predict able to deal with?",
    ),
    (
        "y_inner_mtype",
        "regressor_proba",
        ("list", "str"),
        "which machine type(s) is the internal _fit/_predict able to deal with?",
    ),
    (
        "C_inner_mtype",
        "regressor_proba",
        ("list", "str"),
        "which machine type(s) is the internal _fit/_predict able to deal with?",
    ),
    # ----------------
    # BaseDistribution
    # ----------------
    (
        "capabilities:approx",
        "distribution",
        ("list", "str"),
        "methods of distr that are approximate",
    ),
    (
        "capabilities:exact",
        "distribution",
        ("list", "str"),
        "methods of distr that are numerically exact",
    ),
    (
        "distr:measuretype",
        "distribution",
        ("str", ["continuous", "discrete", "mixed"]),
        "measure type of distr",
    ),
    (
        "distr:paramtype",
        "distribution",
        ("str", ["general", "parametric", "nonparametric", "composite"]),
        "parametrization type of distribution",
    ),
    (
        "approx_mean_spl",
        "distribution",
        "int",
        "sample size used in MC estimates of mean",
    ),
    (
        "approx_var_spl",
        "distribution",
        "int",
        "sample size used in MC estimates of var",
    ),
    (
        "approx_energy_spl",
        "distribution",
        "int",
        "sample size used in MC estimates of energy",
    ),
    (
        "approx_spl",
        "distribution",
        "int",
        "sample size used in other MC estimates",
    ),
    (
        "bisect_iter",
        "distribution",
        "int",
        "max iters for bisection method in ppf",
    ),
    (
        "broadcast_params",
        "distribution",
        ("list", "str"),
        "distribution parameters to broadcast, complement is not broadcast",
    ),
    (
        "broadcast_init",
        "distribution",
        ("str", ["on", "off"]),
        "whether to initialize broadcast parameters in __init__, 'on' or 'off'",
    ),
    (
        "broadcast_inner",
        "distribution",
        ("str", ["array", "scalar"]),
        "if inner logic is vectorized ('array') or scalar ('scalar')",
    ),
    # ---------------
    # BaseProbaMetric
    # ---------------
    (
        "scitype:y_pred",
        "metric",
        "str",
        "expected input type for y_pred in performance metric",
    ),
    (
        "lower_is_better",
        "metric",
        "bool",
        "whether lower (True) or higher (False) is better",
    ),
    (
        "capability:survival",
        "metric",
        "bool",
        "whether metric uses censoring information, for survival analysis",
    ),
    # ----------------------------
    # BaseMetaObject reserved tags
    # ----------------------------
    (
        "named_object_parameters",
        "object",
        "str",
        "name of component list attribute for meta-objects",
    ),
    (
        "fitted_named_object_parameters",
        "estimator",
        "str",
        "name of fitted component list attribute for meta-objects",
    ),
]

OBJECT_TAG_TABLE = pd.DataFrame(OBJECT_TAG_REGISTER)
OBJECT_TAG_LIST = OBJECT_TAG_TABLE[0].tolist()


def check_tag_is_valid(tag_name, tag_value):
    """Check validity of a tag value.

    Parameters
    ----------
    tag_name : string, name of the tag
    tag_value : object, value of the tag

    Raises
    ------
    KeyError - if tag_name is not a valid tag in OBJECT_TAG_LIST
    ValueError - if the tag_valid is not a valid for the tag with name tag_name
    """
    if tag_name not in OBJECT_TAG_LIST:
        raise KeyError(tag_name + " is not a valid tag")

    tag_type = OBJECT_TAG_TABLE[2][OBJECT_TAG_TABLE[0] == "tag_name"]

    if tag_type == "bool" and not isinstance(tag_value, bool):
        raise ValueError(tag_name + " must be True/False, found " + tag_value)

    if tag_type == "int" and not isinstance(tag_value, int):
        raise ValueError(tag_name + " must be integer, found " + tag_value)

    if tag_type == "str" and not isinstance(tag_value, str):
        raise ValueError(tag_name + " must be string, found " + tag_value)

    if tag_type == "list" and not isinstance(tag_value, list):
        raise ValueError(tag_name + " must be list, found " + tag_value)

    if tag_type[0] == "str" and tag_value not in tag_type[1]:
        raise ValueError(
            tag_name + " must be one of " + tag_type[1] + " found " + tag_value
        )

    if tag_type[0] == "list" and not set(tag_value).issubset(tag_type[1]):
        raise ValueError(
            tag_name + " must be subest of " + tag_type[1] + " found " + tag_value
        )

    if tag_type[0] == "list" and tag_type[1] == "str":
        msg = f"{tag_name} must be str or list of str, found {tag_value}"
        if not isinstance(tag_value, (str, list)):
            raise ValueError(msg)
        if isinstance(tag_value, list):
            if not all(isinstance(x, str) for x in tag_value):
                raise ValueError(msg)



"""Test configs."""

# list of str, names of estimators to exclude from testing
# WARNING: tests for these estimators will be skipped
EXCLUDE_ESTIMATORS = [
    "DummySkipped",
    "ClassName",  # exclude classes from extension templates
]


EXCLUDED_TESTS = {"GLMRegressor": ["test_online_update"]}



"""Adapters to ngboost regressors with probabilistic components."""
# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)

__author__ = ["ShreeshaM07"]

from skpro.regression.adapters.ngboost._ngboost_proba import NGBoostAdapter
from skpro.regression.base import BaseProbaRegressor


class NGBoostRegressor(BaseProbaRegressor, NGBoostAdapter):
    """Natural Gradient Boosting Regressor for probabilistic regressors.

    It is an interface to the NGBRegressor.
    NGBRegressor is a wrapper for the generic NGBoost class that facilitates regression.
    Use this class if you want to predict an outcome that could take an
    infinite number of (ordered) values.

    Parameters
    ----------
    dist : string , default = "Normal"
        distribution that must be used for
        probabilistic prediction.
        Available distribution types

        1. "Normal"
        2. "Laplace"
        3. "LogNormal"
        4. "Poisson"
        5. "TDistribution"
        6. "Exponential"

    score : string , default = "LogScore"
        A score from ngboost.scores for LogScore
        rule to compare probabilistic
        predictions P̂ to the observed data y.

    estimator : default learner/estimator: DecisionTreeRegressor()
        base learner to use in the boosting algorithm.
        Any instantiated sklearn regressor.

    natural_gradient : boolean , default = True
        whether natural gradient must be used or not.
    n_estimators : int , default = 500
        the number of boosting iterations to fit
    learning_rate : float , default = 0.01
        the learning rate
    minibatch_frac : float, default = 1.0
        the percent subsample of rows to
        use in each boosting iteration
    verbose : boolean, default=True
        flag indicating whether output
        should be printed during fitting
    verbose_eval : int ,default=100
        increment (in boosting iterations) at
        which output should be printed
    tol : float, default = 1e-4
        numerical tolerance to be used in optimization
    random_state : int, RandomState instance or None, optional (default=None)
    validation_fraction : Proportion of training data to
        set aside as validation data for early stopping.
    early_stopping_rounds : int , default = None , optional
        The number of consecutive
        boosting iterations during which the
        loss has to increase before the algorithm stops early.
        Set to None to disable early stopping and validation
        None enables running over the full data set.

    Example
    -------
    from sklearn.datasets import load_diabetes
    from sklearn.linear_model import LinearRegression
    from sklearn.model_selection import train_test_split
    from skpro.regression.ensemble import NGBoostRegressor
    from skpro.metrics import LogLoss

    X, y = load_diabetes(return_X_y=True, as_frame=True)
    X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2)
    ngb = NGBoostRegressor(
        dist="Normal",
        score="LogScore",
        estimator=LinearRegression(),
        n_estimators=100,
    )
    ngb.fit(X_train, y_train)
    y_pred = ngb.predict_proba(X_test)

    logloss = LogLoss()
    score = logloss(y_test, y_pred)
    """

    _tags = {
        "authors": ["ShreeshaM07"],
        "maintainers": ["ShreeshaM07"],
        "python_dependencies": "ngboost",
    }

    def __init__(
        self,
        dist="Normal",
        score="LogScore",
        estimator=None,
        natural_gradient=True,
        n_estimators=500,
        learning_rate=0.01,
        minibatch_frac=1.0,
        col_sample=1.0,
        verbose=True,
        verbose_eval=100,
        tol=1e-4,
        random_state=None,
        validation_fraction=0.1,
        early_stopping_rounds=None,
    ):
        self.dist = dist
        self.score = score
        self.estimator = estimator
        self.natural_gradient = natural_gradient
        self.n_estimators = n_estimators
        self.learning_rate = learning_rate
        self.minibatch_frac = minibatch_frac
        self.col_sample = col_sample
        self.verbose = verbose
        self.verbose_eval = verbose_eval
        self.tol = tol
        self.random_state = random_state
        self.validation_fraction = validation_fraction
        self.early_stopping_rounds = early_stopping_rounds

        super().__init__()

    def _fit(self, X, y):
        """Fit regressor to training data.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        from ngboost import NGBRegressor
        from ngboost.scores import LogScore
        from sklearn.base import clone
        from sklearn.tree import DecisionTreeRegressor

        # coerce y to numpy array
        y = self._check_y(y=y)
        y = y[0]
        # remember y columns to predict_proba
        self._y_cols = y.columns
        y = y.values.ravel()

        if self.estimator is None:
            self.estimator_ = DecisionTreeRegressor(
                criterion="friedman_mse",
                min_samples_split=2,
                min_samples_leaf=1,
                min_weight_fraction_leaf=0.0,
                max_depth=3,
                splitter="best",
                random_state=None,
            )
        else:
            self.estimator_ = clone(self.estimator)

        dist_ngboost = self._dist_to_ngboost_instance(self.dist, survival=False)

        # Score argument for NGBRegressor
        ngboost_score = {
            "LogScore": LogScore,
        }
        score = None
        if self.score in ngboost_score:
            score = ngboost_score[self.score]

        self.ngb = NGBRegressor(
            Dist=dist_ngboost,
            Score=score,
            Base=self.estimator_,
            natural_gradient=True,
            n_estimators=self.n_estimators,
            learning_rate=self.learning_rate,
            minibatch_frac=self.minibatch_frac,
            col_sample=self.col_sample,
            verbose=self.verbose,
            verbose_eval=self.verbose_eval,
            tol=self.tol,
            random_state=self.random_state,
            validation_fraction=self.validation_fraction,
            early_stopping_rounds=self.early_stopping_rounds,
        )

        self.ngb_ = clone(self.ngb)
        self.ngb_.fit(X, y)
        return self

    def _predict(self, X):
        """Predict labels for data from features.

        State required:
            Requires state to be "fitted" = self.is_fitted=True

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : pandas DataFrame, same length as `X`, same columns as `y` in `fit`
            labels predicted for `X`
        """
        import pandas as pd

        df = pd.DataFrame(self.ngb_.predict(X), index=X.index, columns=self._y_cols)
        return df

    def _pred_dist(self, X):
        return self.ngb_.pred_dist(X)

    def _predict_proba(self, X):
        """Predict distribution over labels for data from features.

        State required:
            Requires state to be "fitted".

        Accesses in self:
            Fitted model attributes ending in "_"

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        y : skpro BaseDistribution, same length as `X`
            labels predicted for `X`
        """
        X = self._check_X(X)

        kwargs = {}
        pred_dist = self._pred_dist(X)
        index = X.index
        columns = self._y_cols

        # Convert NGBoost Distribution return params into a dict
        kwargs = self._ngb_skpro_dist_params(pred_dist, index, columns, **kwargs)

        # Convert NGBoost Distribution to skpro BaseDistribution
        pred_dist = self._ngb_dist_to_skpro(**kwargs)

        return pred_dist

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        from sklearn.ensemble import RandomForestRegressor
        from sklearn.linear_model import LinearRegression

        params1 = {"dist": "Normal"}
        params2 = {
            "dist": "Laplace",
            "n_estimators": 800,
            "estimator": LinearRegression(),
        }
        params3 = {}
        params4 = {
            "dist": "Poisson",
            "minibatch_frac": 0.8,
            "early_stopping_rounds": 4,
        }
        params5 = {
            "dist": "LogNormal",
            "learning_rate": 0.001,
            "validation_fraction": 0.2,
            "estimator": RandomForestRegressor(),
        }

        params6 = {
            "dist": "Normal",
            "natural_gradient": False,
            "verbose": False,
        }

        params7 = {
            "dist": "Exponential",
            "n_estimators": 800,
            "verbose_eval": 50,
        }

        return [params1, params2, params3, params4, params5, params6, params7]



"""Survival or time-to-event prediction estimators, composers."""

from skpro.regression.compose import Pipeline
from skpro.survival.compose._reduce_cond_unc import ConditionUncensored
from skpro.survival.compose._reduce_uncensored import FitUncensored

__all__ = [
    "Pipeline",
    "FitUncensored",
    "ConditionUncensored",
]



# copyright: skpro developers, BSD-3-Clause License (see LICENSE file)
"""Empirical distribution."""

__author__ = ["fkiraly"]

import numpy as np
import pandas as pd

from skpro.distributions.base import BaseDistribution


class Empirical(BaseDistribution):
    r"""Empirical distribution, or weighted sum of delta distributions.

    This distribution represents an empirical distribution, or, more generally,
    a weighted sum of delta distributions.

    The distribution is parameterized by support in ``spl``, and optionally
    weights in ``weights``.

    For the scalar case, the distribution is parameterized as follows:
    let :math:`s_i, i = 1 \dots N` the entries of ``spl``,
    and :math:`w_i, i = 1 \dots N` the entries of ``weights``; if ``weights=None``,
    by default we define :math:`p_i = \frac{1}{N}`, otherwise we
    define :math:`p_i := \frac{w_i}{\sum_{i=1}^N w_i}`

    The distribution is the unique distribution that takes value :math:`s_i` with
    probability :math:`p_i`. In particluar, if ``weights`` was ``None``,
    the distribution is the uniform distribution supported on the :math:`s_i`.

    Parameters
    ----------
    spl : pd.DataFrame
        empirical sample; for scalar distributions, rows are samples;
        for dataframe-like distributions,
        first (lowest) index is sample, further indices are instance indices
    weights : pd.Series, with same index and length as spl, optional, default=None
        if not passed, ``spl`` is assumed to be unweighted
    time_indep : bool, optional, default=True
        if True, ``sample`` will sample individual instance indices independently
        if False, ``sample`` will sample entire instances from ``spl``
    index : pd.Index, optional, default = RangeIndex
    columns : pd.Index, optional, default = RangeIndex

    Examples
    --------
    >>> import pandas as pd
    >>> from skpro.distributions.empirical import Empirical

    >>> spl_idx = pd.MultiIndex.from_product(
    ...     [[0, 1], [0, 1, 2]], names=["sample", "time"]
    ... )
    >>> spl = pd.DataFrame(
    ...     [[0, 1], [2, 3], [10, 11], [6, 7], [8, 9], [4, 5]],
    ...     index=spl_idx,
    ...     columns=["a", "b"],
    ... )
    >>> dist = Empirical(spl)
    >>> empirical_sample = dist.sample(3)

    scalar distribution:
    >>> spl = pd.Series([1, 2, 3, 4, 3])
    >>> dist = Empirical(spl)
    >>> empirical_sample = dist.sample(3)
    """

    _tags = {
        "capabilities:approx": [],
        "capabilities:exact": ["mean", "var", "energy", "cdf", "ppf"],
        "distr:measuretype": "discrete",
        "distr:paramtype": "nonparametric",
    }

    def __init__(self, spl, weights=None, time_indep=True, index=None, columns=None):
        self.spl = spl
        self.weights = weights
        self.time_indep = time_indep

        index, columns = self._init_index(index, columns)

        super().__init__(index=index, columns=columns)

        # initialized sorted samples
        self._init_sorted()

    def _init_index(self, index, columns):
        """Initialize index and columns.

        Sets the following attributes:

        * ``_spl_indices`` - unique index for samples
        * ``_shape`` - shape of self - 0D or 2D
        * ``_N`` - number of samples
        * only if array distribution: ``_instances``,
          coerced index of ``self``, from ``spl`` index
        """
        spl = self.spl

        is_scalar = not isinstance(spl.index, pd.MultiIndex)
        is_scalar = is_scalar and (spl.ndim <= 1 or spl.shape[1] == 1)

        if is_scalar:
            self._shape = ()
            _spl_indices = spl.index
            self._spl_indices = _spl_indices
            self._N = len(_spl_indices)
            return None, None

        _instances = spl.index.droplevel(0).unique()
        _spl_indices = spl.index.get_level_values(0).unique()
        self._instances = _instances
        self._spl_indices = _spl_indices
        self._N = len(_spl_indices)

        if index is None:
            index = _instances
        if columns is None:
            columns = spl.columns

        self._shape = (len(index), len(columns))

        return index, columns

    def _init_sorted(self):
        """Initialize sorted version of spl."""
        if self.ndim == 0:
            spl = self.spl.values.flatten()
            sorter = np.argsort(spl)
            spl_sorted = spl[sorter]
            if self.weights is not None:
                weights_sorted = self.weights.values.flatten()[sorter]
            else:
                weights_sorted = np.ones_like(spl)
            self._sorted = spl_sorted
            self._weights = weights_sorted
            return None

        times = self._instances
        cols = self.columns

        sorted = {}
        weights = {}
        for t in times:
            sorted[t] = {}
            weights[t] = {}
            for col in cols:
                sl = (slice(None),) + self._coerce_tuple(t)
                spl_t = self.spl.loc[sl, col].values
                sorter = np.argsort(spl_t)
                spl_t_sorted = spl_t[sorter]
                sorted[t][col] = spl_t_sorted
                if self.weights is not None:
                    weights_t = self.weights.loc[sl].values
                    weights_t_sorted = weights_t[sorter]
                    weights[t][col] = weights_t_sorted
                else:
                    ones = np.ones(len(spl_t_sorted))
                    weights[t][col] = ones

        self._sorted = sorted
        self._weights = weights

    def _coerce_tuple(self, x):
        if not isinstance(x, tuple):
            x = (x,)
        return x

    def _apply_per_ix(self, func, params, x=None):
        """Apply function per index."""
        sorted = self._sorted
        weights = self._weights

        if self.ndim == 0:
            return func(spl=sorted, weights=weights, x=x, **params)

        index = self.index
        cols = self.columns

        res = pd.DataFrame(index=index, columns=cols)
        for i, ix in enumerate(index):
            for j, col in enumerate(cols):
                spl_t = sorted[ix][col]
                weights_t = weights[ix][col]
                if x is None:
                    x_t = None
                elif hasattr(x, "loc"):
                    x_t = x.loc[ix, col]
                else:
                    x_t = x[i, j]
                res.at[ix, col] = func(spl=spl_t, weights=weights_t, x=x_t, **params)
        return res.apply(pd.to_numeric)

    def _slice_ix(self, obj, ix):
        """Slice obj by index ix, applied to MultiIndex levels 1 ... last.

        obj is assumed to have MultiIndex, and slicing occurs on the
        last levels, 1 ... last.

        ix can be a simple index or MultiIndex,
        same number of levels as obj.index.droplevel(0).

        This utility function is needed since pandas cannot do this?

        Parameters
        ----------
        obj : pd.DataFrame or pd.Series, with pd.MultiIndex
            object to slice
        ix : pd.Index or pd.MultiIndex
            index to slice by, loc references

        Returns
        -------
        pd.DataFrame or pd.Series, same type as obj
            obj sliced by levels 1 ... last, subset to levels in ix
        """
        if not isinstance(ix, pd.MultiIndex):
            if isinstance(obj, pd.DataFrame):
                return obj.loc[(slice(None), ix), :]
            else:
                return obj.loc[(slice(None), ix)]

        obj_ix = obj.index
        obj_vals = obj_ix.get_level_values(0).unique()
        if isinstance(ix, pd.MultiIndex):
            prod_ix = [(v,) + i for v in obj_vals for i in ix]
        else:
            prod_ix = [(v,) + (i,) for v in obj_vals for i in ix]
        prod_ix = pd.MultiIndex.from_tuples(prod_ix)

        return obj.loc[prod_ix]

    def _iloc(self, rowidx=None, colidx=None):
        if is_scalar_notnone(rowidx) and is_scalar_notnone(colidx):
            return self._iat(rowidx, colidx)
        if is_scalar_notnone(rowidx):
            rowidx = pd.Index([rowidx])
        if is_scalar_notnone(colidx):
            colidx = pd.Index([colidx])

        index = self.index
        columns = self.columns
        weights = self.weights

        spl_subset = self.spl

        if rowidx is not None:
            rowidx_loc = index[rowidx]
            # subset multiindex to rowidx by last level
            spl_subset = self._slice_ix(self.spl, rowidx_loc)
            if weights is not None:
                weights_subset = self._slice_ix(weights, rowidx_loc)
            else:
                weights_subset = None
            subs_rowidx = index[rowidx]
        else:
            subs_rowidx = index
            weights_subset = weights

        if colidx is not None:
            spl_subset = spl_subset.iloc[:, colidx]
            subs_colidx = columns[colidx]
        else:
            subs_colidx = columns

        return Empirical(
            spl_subset,
            weights=weights_subset,
            time_indep=self.time_indep,
            index=subs_rowidx,
            columns=subs_colidx,
        )

    def _iat(self, rowidx=None, colidx=None):
        if rowidx is None or colidx is None:
            raise ValueError("iat method requires both row and column index")
        self_subset = self.iloc[[rowidx], [colidx]]
        levels_to_drop = list(range(1, self_subset.spl.index.nlevels))
        spl_subset = self_subset.spl.droplevel(levels_to_drop)
        if self.weights is not None:
            wts_subset = self_subset.weights.droplevel(levels_to_drop)
        else:
            wts_subset = None

        subset_params = {"spl": spl_subset, "weights": wts_subset}
        return type(self)(**subset_params)

    def _energy_default(self, x=None):
        r"""Energy of self, w.r.t. self or a constant frame x.

        Let :math:`X, Y` be i.i.d. random variables with the distribution of `self`.

        If `x` is `None`, returns :math:`\mathbb{E}[|X-Y|]` (per row), "self-energy".
        If `x` is passed, returns :math:`\mathbb{E}[|X-x|]` (per row), "energy wrt x".

        Parameters
        ----------
        x : None or pd.DataFrame, optional, default=None
            if pd.DataFrame, must have same rows and columns as `self`

        Returns
        -------
        pd.DataFrame with same rows as `self`, single column `"energy"`
        each row contains one float, self-energy/energy as described above.
        """
        energy_arr = self._apply_per_ix(_energy_np, {"assume_sorted": True}, x=x)
        if energy_arr.ndim > 0:
            energy_arr = np.sum(energy_arr, axis=1)
        return energy_arr

    def _mean(self):
        r"""Return expected value of the distribution.

        Let :math:`X` be a random variable with the distribution of `self`.
        Returns the expectation :math:`\mathbb{E}[X]`

        Returns
        -------
        pd.DataFrame with same rows, columns as `self`
        expected value of distribution (entry-wise)
        """
        spl = self.spl

        # scalar case
        if self.ndim == 0:
            spl = spl.values.flatten()
            if self.weights is None:
                return np.mean(spl)
            else:
                return np.average(spl, weights=self.weights)

        # dataframe case
        groupby_levels = list(range(1, spl.index.nlevels))
        if self.weights is None:
            mean_df = spl.groupby(level=groupby_levels, sort=False).mean()
        else:
            mean_df = spl.groupby(level=groupby_levels, sort=False).apply(
                lambda x: np.average(x, weights=self.weights.loc[x.index], axis=0)
            )
            mean_df = pd.DataFrame(mean_df.tolist(), index=mean_df.index)
            mean_df.columns = spl.columns

        mean_df = mean_df.loc[self.index]  # ensure consistent sorting
        return mean_df

    def _var(self):
        r"""Return element/entry-wise variance of the distribution.

        Let :math:`X` be a random variable with the distribution of `self`.
        Returns :math:`\mathbb{V}[X] = \mathbb{E}\left(X - \mathbb{E}[X]\right)^2`

        Returns
        -------
        pd.DataFrame with same rows, columns as `self`
        variance of distribution (entry-wise)
        """
        spl = self.spl
        N = self._N

        # scalar case
        if self.ndim == 0:
            spl = spl.values.flatten()
            if self.weights is None:
                return np.var(spl, ddof=0)
            else:
                mean = self.mean()
                var = np.average((spl - mean) ** 2, weights=self.weights)
                return var

        # dataframe case
        groupby_levels = list(range(1, spl.index.nlevels))
        if self.weights is None:
            var_df = spl.groupby(level=groupby_levels, sort=False).var(ddof=0)
        else:
            mean = self.mean()
            means = pd.concat([mean] * N, axis=0, keys=self._spl_indices)
            var_df = spl.groupby(level=groupby_levels, sort=False).apply(
                lambda x: np.average(
                    (x - means.loc[x.index]) ** 2,
                    weights=self.weights.loc[x.index],
                    axis=0,
                )
            )
            var_df = pd.DataFrame(
                var_df.tolist(), index=var_df.index, columns=spl.columns
            )

        var_df = var_df.loc[self.index]  # ensure consistent sorting
        return var_df

    def _cdf(self, x):
        """Cumulative distribution function."""
        cdf_val = self._apply_per_ix(_cdf_np, {"assume_sorted": True}, x=x)
        return cdf_val

    def _ppf(self, p):
        """Quantile function = percent point function = inverse cdf."""
        ppf_val = self._apply_per_ix(_ppf_np, {"assume_sorted": True}, x=p)
        return ppf_val

    def sample(self, n_samples=None):
        """Sample from the distribution.

        Parameters
        ----------
        n_samples : int, optional, default = None
            number of samples to draw from the distribution

        Returns
        -------
        pd.DataFrame
            samples from the distribution

            * if ``n_samples`` is ``None``:
            returns a sample that contains a single sample from ``self``,
            in ``pd.DataFrame`` mtype format convention, with ``index`` and ``columns``
            as ``self``
            * if n_samples is ``int``:
            returns a ``pd.DataFrame`` that contains ``n_samples`` i.i.d.
            samples from ``self``, in ``pd-multiindex`` mtype format convention,
            with same ``columns`` as ``self``, and row ``MultiIndex`` that is product
            of ``RangeIndex(n_samples)`` and ``self.index``
        """
        # for now, always defaulting to the standard logic
        # todo: address issue #283
        if self.ndim >= 0:
            return super().sample(n_samples=n_samples)

        spl = self.spl
        timestamps = self._instances
        weights = self.weights

        if n_samples is None:
            n_samples = 1
            n_samples_was_none = True
        else:
            n_samples_was_none = False
        smpls = []

        for _ in range(n_samples):
            smpls_i = []
            for t in timestamps:
                spl_from = spl.loc[(slice(None), t), :]
                if weights is not None:
                    spl_weights = weights.loc[(slice(None), t)].values
                else:
                    spl_weights = None
                spl_time = spl_from.sample(n=1, replace=True, weights=spl_weights)
                spl_time = spl_time.droplevel(0)
                smpls_i.append(spl_time)
            spl_i = pd.concat(smpls_i, axis=0)
            smpls.append(spl_i)

        spl = pd.concat(smpls, axis=0, keys=range(n_samples))
        if n_samples_was_none:
            spl = spl.droplevel(0)

        return spl

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator."""
        # params1 is a DataFrame with simple row multiindex
        spl_idx = pd.MultiIndex.from_product(
            [[0, 1], [0, 1, 2]], names=["sample", "time"]
        )
        spl = pd.DataFrame(
            [[0, 1], [2, 3], [10, 11], [6, 7], [8, 9], [4, 5]],
            index=spl_idx,
            columns=["a", "b"],
        )
        params1 = {
            "spl": spl,
            "weights": None,
            "time_indep": True,
            "index": pd.RangeIndex(3),
            "columns": pd.Index(["a", "b"]),
        }

        # params2 is weighted
        params2 = {
            "spl": spl,
            "weights": pd.Series([0.5, 0.5, 0.5, 1, 1, 1.1], index=spl_idx),
            "time_indep": False,
            "index": pd.RangeIndex(3),
            "columns": pd.Index(["a", "b"]),
        }

        # params3 is scalar, unweighted
        spl_scalar = pd.Series([1, 2, 3, 4, 3])
        params3 = {"spl": spl_scalar}

        # params4 is scalar, weighted
        spl_scalar = pd.DataFrame([1, 2, 3, 4, 3])
        wts_scalar = pd.Series([0.2, 0.2, 0.3, 0.3, 0.1])
        params4 = {"spl": spl_scalar, "weights": wts_scalar}

        # hierarchical MultiIndex, important for sktime
        spl_idx = pd.MultiIndex.from_product(
            [[0, 1], [0, 1, 2], [0, 1]], names=["sample", "instance", "time"]
        )
        param5 = {"spl": pd.DataFrame(np.random.rand(12, 2), index=spl_idx)}

        # hierarchical MultiIndex, important for sktime, weighted
        spl_idx = pd.MultiIndex.from_product(
            [[0, 1], [0, 1, 2], [0, 1]], names=["sample", "instance", "time"]
        )
        weights = pd.Series(list(range(1, 13)), index=spl_idx)
        spl = pd.DataFrame(np.random.rand(12, 2), index=spl_idx)
        param6 = {"spl": spl, "weights": weights}

        return [params1, params2, params3, params4, param5, param6]


def _energy_np(spl, x=None, weights=None, assume_sorted=False):
    r"""Compute sample energy, fast numpy based subroutine.

    Let :math:`X` be the random variable with support being
    values of `spl`, with probability weights `weights`.

    This function then returns :math:`\mathbb{E}[|X-Y|]`, with :math:`Y` an
    independent copy of :math:`X`, if `x` is `None`.

    If `x` is passed, returns :math:`\mathbb{E}[|X-x|]`.

    Parameters
    ----------
    spl : 1D np.ndarray
        empirical sample
    x : None or float, optional, default=None
        if None, computes self-energy, if float, computes energy wrt x
    weights : None or 1D np.ndarray, optional, default=None
        if None, computes unweighted energy, if 1D np.ndarray, computes weighted energy
        if not None, must be of same length as ``spl``, needs not be normalized
    assume_sorted : bool, optional, default=False
        if True, assumes that ``spl`` is sorted in ascending order

    Returns
    -------
    float
        energy as described above
    """
    if weights is None:
        weights = np.ones(len(spl))

    if not assume_sorted:
        sorter = np.argsort(spl)
        spl = spl[sorter]
        weights = weights[sorter]

    w_sum = np.sum(weights)
    weights = weights / w_sum

    spl_diff = np.diff(spl)

    if x is None:
        cum_fwd = np.cumsum(weights[:-1])
        cum_back = np.cumsum(weights[1:][::-1])[::-1]
        energy = 2 * np.sum(cum_fwd * cum_back * spl_diff)
    else:
        spl_diff = np.abs(spl - x)
        energy = np.sum(weights * spl_diff)

    return energy


def _cdf_np(spl, x, weights=None, assume_sorted=False):
    """Compute empirical cdf, fast numpy based subroutine.

    Parameters
    ----------
    spl : 1D np.ndarray
        empirical sample
    x : float
        value at which to evaluate cdf
    weights : None or 1D np.ndarray, optional, default=None
        if None, computes unweighted cdf, if 1D np.ndarray, computes weighted cdf
        if not None, must be of same length as ``spl``, needs not be normalized
    assume_sorted : bool, optional, default=False
        if True, assumes that ``spl`` is sorted in ascending order

    Returns
    -------
    cdf_val float
        cdf-value at x
    """
    if weights is None:
        weights = np.ones(len(spl))

    if not assume_sorted:
        sorter = np.argsort(spl)
        spl = spl[sorter]
        weights = weights[sorter]

    w_sum = np.sum(weights)
    weights = weights / w_sum

    weights_select = weights[spl <= x]
    cdf_val = np.sum(weights_select)

    return cdf_val


def _ppf_np(spl, x, weights=None, assume_sorted=False):
    """Compute empirical ppf, fast numpy based subroutine.

    Parameters
    ----------
    spl : 1D np.ndarray
        empirical sample
    x : float
        probability at which to evaluate ppf
    weights : None or 1D np.ndarray, optional, default=None
        if None, computes unweighted ppf, if 1D np.ndarray, computes weighted ppf
        if not None, must be of same length as ``spl``, needs not be normalized
    assume_sorted : bool, optional, default=False
        if True, assumes that ``spl`` is sorted in ascending order

    Returns
    -------
    ppf_val float
        ppf-value at p
    """
    if weights is None:
        weights = np.ones(len(spl))

    if not assume_sorted:
        sorter = np.argsort(spl)
        spl = spl[sorter]
        weights = weights[sorter]

    w_sum = np.sum(weights)
    weights = weights / w_sum

    cum_weights = np.cumsum(weights)
    ix_val = np.searchsorted(cum_weights, x)
    ppf_val = spl[ix_val]

    return ppf_val


def is_scalar_notnone(obj):
    """Check if obj is scalar and not None."""
    return obj is not None and np.isscalar(obj)



# LEGACY MODULE - TODO: remove or refactor

if False:
    import numpy as np
    from uncertainties import ufloat
    from sklearn.model_selection import GridSearchCV, RandomizedSearchCV
    from sklearn.model_selection import KFold

    from ..model_selection import cross_val_score
    from ..metrics import make_scorer

from .base import Controller, View


def grid_optimizer(verbose=0, n_jobs=1):
    def wrapper(model, search_space, scoring, cv):
        return GridSearchCV(
            estimator=model,
            param_grid=search_space,
            scoring=scoring,
            cv=cv,
            verbose=verbose,
            n_jobs=n_jobs,
        )

    return wrapper


def random_optimizer(n_iter=10, verbose=0, n_jobs=1):
    def wrapper(model, search_space, scoring, cv):
        return RandomizedSearchCV(
            estimator=model,
            param_distributions=search_space,
            scoring=scoring,
            cv=cv,
            n_iter=n_iter,
            verbose=verbose,
            n_jobs=n_jobs,
        )

    return wrapper


class CrossValidationController(Controller):
    """CrossValidation controller

    Parameters
    ----------
    data
    loss_func
    cv
    tune
    optimizer: optional callable(model, search_space, scoring, cv), default=None
        Callable that returns an optimizer such as RandomizedSearchCV etc. that
        will be used for the hyperparameter search.
    """

    def __init__(self, data, loss_func, cv=None, tune=False, optimizer=None):
        self.data = data
        self.loss_func = loss_func
        if cv is None:
            cv = KFold()
        self.cv = cv
        self.tune = tune
        if optimizer is None:
            optimizer = grid_optimizer()
        self.optimizer = optimizer

    def identifier(self):
        return "CrossValidation(data=%s, loss_func=%s)" % (
            self.data.name,
            self.loss_func.__name__,
        )

    def __repr__(self):
        return "CrossValidation(data=%s, loss_func=%s, cv=%s, tune=%s)" % (
            self.data.name,
            self.loss_func.__name__,
            repr(self.cv),
            str(self.tune),
        )

    def description(self):
        return "CV(%s, %s)" % (self.data.name, self.loss_func.__name__)

    def run(self, model):

        if self.tune:
            tuning = model.tuning
        else:
            tuning = None

        if tuning is None:
            clf = model.instance
            best_params = None
            scorer = make_scorer(self.loss_func)
        else:
            clf = self.optimizer(
                model=model.instance,
                search_space=tuning,
                scoring=make_scorer(self.loss_func, greater_is_better=False),
                cv=self.cv,
            )

            best_params = []
            loss_function = self.loss_func

            def scorer(estimator, X_test, y_test, **kwargs):
                y_pred = estimator.predict(X_test)
                best_params.append(estimator.best_params_)
                return loss_function(y_test, y_pred, **kwargs)

        scores = cross_val_score(
            clf, self.data.X, self.data.y, scoring=scorer, cv=self.cv
        )

        score = ufloat(np.mean(scores[:, 0]), np.mean(scores[:, 1]))

        return {
            "score": score,
            "scores": scores,
            "tuning": tuning,
            "best_params": best_params,
        }


class CrossValidationView(View):
    """Cross validation view

    Parameters
    ----------
    with_tuning
    with_ranks
    """

    def __init__(self, with_tuning=False, with_ranks=True):
        self.with_tuning = with_tuning
        self.with_ranks = with_ranks

    def parse(self, data):
        result = str(data["score"])

        if data["tuning"] is not None:
            if self.with_tuning:
                best = data["best_params"][0]
                tuning = ""
                comma = ""
                for k, v in data["tuning"].items():
                    tuning += comma + k + ": "
                    comma = "; "

                    sep = ""
                    for e in v:
                        t = "%s"
                        if e == best[k]:
                            t = "**%s**"

                        tuning += sep + (t % e)
                        sep = ", "

                result += "* (%s)" % tuning
            else:
                result += "*"

        if self.with_ranks and "vrank" in data:
            result = "(%i) " % data["vrank"] + result

        return result



"""Dummy time series regressor."""

__author__ = ["julian-fong"]
__all__ = ["DummyProbaRegressor"]

import numpy as np
import pandas as pd

from skpro.distributions.empirical import Empirical
from skpro.distributions.normal import Normal
from skpro.regression.base import BaseProbaRegressor


class DummyProbaRegressor(BaseProbaRegressor):
    """DummyProbaRegressor makes predictions that ignore the input features.

    This regressor serves as a simple baseline to compare against other more
    complex regressors.
    The specific behavior of the baseline is selected with the ``strategy``
    parameter.

    All strategies make predictions that ignore the input feature values passed
    as the ``X`` argument to ``fit`` and ``predict``. The predictions, however,
    typically depend on values observed in the ``y`` parameter passed to ``fit``.

    Parameters
    ----------
    strategy : one of ["empirical", "normal"] default="empirical"
        Strategy to use to generate predictions.

        * "empirical": always predicts the empirical unweighted distribution
            of the training labels
        * "normal": always predicts a normal distribution, with mean and variance
            equal to the mean and variance of the training labels

    Attributes
    ----------
    distribution_ : skpro.distribution
        Normal distribution or Empirical distribution, depending on chosen strategy.
        Scalar version of the distribution that is returned by ``predict_proba``.
    """

    _tags = {
        # packaging info
        # --------------
        "authors": ["julian-fong"],
        "maintainers": ["julian-fong"],
        # estimator tags
        # --------------
        "capability:multioutput": False,
        "capability:missing": True,
        "X_inner_mtype": "pd_DataFrame_Table",
        "y_inner_mtype": "pd_DataFrame_Table",
    }

    def __init__(self, strategy="empirical"):
        self.strategy = strategy
        super().__init__()

    def _fit(self, X, y):
        """Fit the dummy regressor.

        Writes to self:
            Sets fitted model attributes ending in "_".

        Parameters
        ----------
        X : pandas DataFrame
            feature instances to fit regressor to
        y : pandas DataFrame, must be same length as X
            labels to fit regressor to

        Returns
        -------
        self : reference to self
        """
        self._y = y
        self._y_columns = y.columns
        self._mu = np.mean(y.values)
        self._sigma = np.std(y.values)
        if self.strategy == "empirical":
            self.distribution_ = Empirical(y)
        if self.strategy == "normal":
            self.distribution_ = Normal(self._mu, self._sigma)

        return self

    def _predict(self, X):
        """Predict labels for data from features.

        Parameters
        ----------
        X : sktime-format pandas dataframe or array-like, shape (n, d)

        Returns
        -------
        y : pandas DataFrame
            predictions of target values for X
        """
        X_ind = X.index
        X_n_rows = X.shape[0]
        y_pred = pd.DataFrame(
            np.ones(X_n_rows) * self._mu, index=X_ind, columns=self._y_columns
        )
        return y_pred

    def _predict_var(self, X):
        """Compute/return variance predictions.

        private _predict_var containing the core logic, called from predict_var

        Parameters
        ----------
        X : pandas DataFrame, must have same columns as X in `fit`
            data to predict labels for

        Returns
        -------
        pred_var : pd.DataFrame
            Column names are exactly those of ``y`` passed in ``fit``.
            Row index is equal to row index of ``X``.
            Entries are variance prediction, for var in col index.
            A variance prediction for given variable and fh index is a predicted
            variance for that variable and index, given observed data.
        """
        X_ind = X.index
        X_n_rows = X.shape[0]
        y_pred = pd.DataFrame(
            np.ones(X_n_rows) * self._sigma, index=X_ind, columns=self._y_columns
        )
        return y_pred

    def _predict_proba(self, X):
        """Broadcast skpro distribution from fit onto labels from X.

        Parameters
        ----------
        X : sktime-format pandas dataframe or array-like, shape (n, d)

        Returns
        -------
        y : skpro.distribution, same length as `X`
            labels predicted for `X`
        """
        X_ind = X.index
        X_n_rows = X.shape[0]
        if self.strategy == "normal":
            # broadcast the mu and sigma from fit to the length of X
            mu = np.reshape((np.ones(X_n_rows) * self._mu), (-1, 1))
            sigma = np.reshape((np.ones(X_n_rows) * self._sigma), (-1, 1))
            pred_dist = Normal(mu=mu, sigma=sigma, index=X_ind, columns=self._y_columns)
            return pred_dist

        if self.strategy == "empirical":
            empr_df = pd.concat([self._y] * X_n_rows, keys=X_ind).swaplevel()
            pred_dist = Empirical(empr_df, index=X_ind, columns=self._y_columns)

            return pred_dist

    @classmethod
    def get_test_params(cls, parameter_set="default"):
        """Return testing parameter settings for the estimator.

        Parameters
        ----------
        parameter_set : str, default="default"
            Name of the set of test parameters to return, for use in tests. If no
            special parameters are defined for a value, will return `"default"` set.

        Returns
        -------
        params : dict or list of dict, default = {}
            Parameters to create testing instances of the class
            Each dict are parameters to construct an "interesting" test instance, i.e.,
            `MyClass(**params)` or `MyClass(**params[i])` creates a valid test instance.
            `create_test_instance` uses the first (or only) dictionary in `params`
        """
        params1 = {}
        params2 = {"strategy": "normal"}

        return [params1, params2]
